"""
Creating a file/directory/item selector component
"""

import json
import pandas as pd

from fusion_tools.visualization.vis_utils import get_pattern_matching_value

# Dash imports
from dash import dcc, callback, ctx, ALL, MATCH, exceptions, no_update, dash_table
import dash_bootstrap_components as dbc
from dash_extensions.enrich import DashBlueprint, html, Input, Output, State, PrefixIdTransform, MultiplexerTransform

from typing_extensions import Union
from fusion_tools.components.base import DSATool


class DSAResourceSelector(DSATool):
    """A small embeddable tool for selecting different folders/slides in a connected DSA instance
    """

    title = 'DSA Resource Selector'
    description = ''

    def __init__(self,
                 handler,
                 selector_type:str = 'item',
                 select_count: Union[int,None] = None
                 ):
        
        super().__init__()

        self.handler = handler
        self.selector_type = selector_type
        self.select_count = select_count


    def get_collection_table(self, session_data:dict, return_data:bool):
        
        user_external_token = self.get_user_external_token(session_data)
        user_external_id = self.get_user_external_id(session_data)

        # Getting user folders if a user is present
        if not user_external_id is None:
            user_folders = [
                "Private",
                "Public"
            ]
        else:
            user_folders = []

        # Getting collection names
        self.handler.gc.setToken(user_external_token)
        collection_list = self.handler.get_collections(user_token = user_external_token)

        if len(user_folders)>0:
            collection_list+=[{'name': 'User Folders','_modelType': 'user', '_id': user_external_id}]

        if not return_data:
            collection_table = html.Div(
                id = {'type': 'dsa-resource-selector-folder-div','index': 0},
                children = [
                    dbc.Stack([
                        html.H5(
                            id = {'type': 'dsa-resource-selector-current-resource-path','index': 0},
                            children = [
                                html.A(
                                    'Collections and User Folders',
                                    id = {'type': 'dsa-resource-selector-path-part','index': 0},
                                    style = {'color': 'rgb(0,0,255)'}
                                )
                            ]
                        ),
                        html.Div(
                            self.make_selectable_dash_table(
                                dataframe = pd.DataFrame.from_records([{'Name': i['name'], 'Type': i['_modelType'], '_id': i['_id']} for i in collection_list]),
                                id = {'type': 'dsa-resource-selector-resource-table','index': 0},
                                selected_rows = [],
                                use_prefix=False
                            ),
                            id = {'type': 'dsa-resource-selector-resource-table-div','index': 0}
                        )
                    ])
                ]
            )

            return collection_table
        else:
            return [{'Name': i['name'], 'Type': i['_modelType'], '_id': i['_id']} for i in collection_list]

    def update_layout(self, session_data: dict, use_prefix:bool, selector_type:str = 'item', select_count: Union[int,None]=None, source_channel: Union[str,None]= None, source_index:Union[int,dict]=0):


        if selector_type in ['image','item']:
            selector_type = 'file'
        self.selector_type = selector_type
        self.select_count = select_count
        self.selector_channel = source_channel

        self.multi_collection = False
        self.multi_folder = False
        self.multi_item = False
        self.multi_file = False
        self.multi_annotation = False

        if self.selector_type in ['item','annotation','file']:
            if self.select_count is None:
                if self.selector_type=='item':
                    self.multi_item = True
                elif self.selector_type=='file':
                    self.multi_file = True
                elif self.selector_type=='annotation':
                    self.multi_annotation = True
        
        if self.selector_type in ['folder']:
            if self.select_count is None:
                self.multi_folder = True
            
        if self.selector_type in ['collection']:
            if self.select_count is None:
                self.multi_collection = True

        collection_table = self.get_collection_table(session_data,return_data=False)

        layout = html.Div([
            html.H4(
                id = {'type': 'dsa-resource-selector-current-user','index': 0},
                children = [
                    f'Showing available resources for: {session_data["user"]["login"]}' if "user" in session_data else "Showing only publically available resources"
                ]
            ),
            html.Hr(),
            dcc.Store(
                id = {'type': 'anchor-dsa-resource-selector-selected-resources','index': 0},
                data = json.dumps({'resource_list':[],'source_component_index': source_index}),
                storage_type='memory'
            ),
            html.Div(
                id = {'type': 'dsa-resource-selector-div','index': 0},
                children = [
                    dbc.Stack([
                        collection_table
                    ])
                ]
            )
        ])

        if use_prefix:
            PrefixIdTransform(prefix = f'{self.component_prefix}', escape = lambda input_id: self.prefix_escape(input_id)).transform_layout(layout)

        return layout

    def make_selectable_dash_table(self, dataframe:pd.DataFrame, id:dict, selected_rows: list = [],use_prefix:bool=True):
        """Generate a selectable DataTable to add to the layout

        :param dataframe: Pandas DataFrame containing columns/rows of interest
        :type dataframe: pd.DataFrame
        :param id: Dictionary containing "type" and "index" keys for interactivity
        :type id: dict
        :param multi_row: Whether to allow selection of multiple rows in the table or just single, defaults to True
        :type multi_row: bool, optional
        :return: dash_table.DataTable component to be added to layout
        :rtype: dash_table.DataTable
        """

        #Optional: Can hide "ID" columns by adding any column containing "ID" to a list of "hidden_columns"
        use_multi = False
        if not dataframe.empty:
            model_types = dataframe['Type'].unique().tolist()
            if len(model_types)==1:
                model_type = model_types[0]
                if model_type=='item' and self.multi_item:
                    use_multi = True
                elif model_type=='file' and self.multi_file:
                    use_multi = True
                elif model_type=='annotation' and self.multi_annotation:
                    use_multi = True
                elif model_type=='folder' and self.multi_folder:
                    use_multi = True         

        dataframe = pd.json_normalize(dataframe.to_dict('records'))
        selectable_table = dash_table.DataTable(
            id = id,
            columns = [{'name':i,'id':i,'deletable':False} for i in dataframe.columns],
            data = dataframe.to_dict('records'),
            editable = False,
            filter_action='native',
            sort_action = 'native',
            sort_mode = 'multi',
            column_selectable = 'single',
            row_selectable = 'multi' if use_multi else 'single',
            row_deletable = False,
            selected_rows = selected_rows,
            page_action='native',
            page_current=0,
            page_size=10,
            style_cell = {
                'overflow':'hidden',
                'textOverflow':'ellipsis',
                'maxWidth':0                
            },
            tooltip_data = [
                {
                    column: {'value':str(value),'type':'markdown'}
                    for column, value in row.items()
                } for row in dataframe.to_dict('records')
            ],
            tooltip_duration = None
        )

        if use_prefix:
            PrefixIdTransform(prefix=f'{self.component_prefix}', escape = lambda input_id: self.prefix_escape(input_id)).transform_layout(selectable_table)

        return selectable_table

    def extract_path_parts(self, current_parts:Union[list,dict], search_key: list = ['props','children'])->tuple:
        """Recursively extract pieces of folder paths stored as clickable components.

        :param current_parts: list or dictionary containing html.A or dbc.Stack of html.A components.
        :type current_parts: Union[list,dict]
        :param search_key: Property keys to search for in nested dicts, defaults to ['props','children']
        :type search_key: list, optional
        :return: Tuple containing all the parts of the folder path
        :rtype: tuple
        """
        path_pieces = ()
        if type(current_parts)==list:
            for c in current_parts:
                if type(c)==dict:
                    for key,value in c.items():
                        if key in search_key:
                            if type(value)==str:
                                path_pieces += (value.replace('/',''),)
                            elif type(value) in [list,dict]:
                                path_pieces += self.extract_path_parts(value)
                elif type(c)==list:
                    path_pieces += self.extract_path_parts(c)
        elif type(current_parts)==dict:
            for key,value in current_parts.items():
                if key in search_key:
                    if type(value)==str:
                        path_pieces += (value.replace('/',''),)
                    elif type(value) in [list,dict]:
                        path_pieces += self.extract_path_parts(value)

        return path_pieces

    def organize_folder_contents(self, folder_info:dict, show_empty:bool=False, ignore_histoqc:bool=True, session_data:dict = {})->list:
        """For a given folder selection, return a list of slides(0th) and folders (1th)

        :param folder_info: Folder info dict returned by self.handler.get_path_info(path)
        :type folder_info: dict
        :param show_empty: Whether or not to display folders which contain 0 slides, defaults to False
        :type show_empty: bool, optional
        :return: List of slides within the current folder as well as folders within that folder
        :rtype: list
        """

        folder_folders = []
        folder_slides = []

        user_external_token = self.get_user_external_token(session_data)

        # Starting with slides (which will report parent folderId but not that parent's folderId (if applicable))
        if folder_info['_modelType'] in ['folder','collection']:
            if folder_info['_modelType']=='folder':
                all_folder_slides = self.handler.get_folder_slides(
                    folder_path = folder_info['_id'],
                    folder_type = folder_info['_modelType'],
                    ignore_histoqc=ignore_histoqc,
                    user_token = user_external_token
                )
            else:
                all_folder_slides = []

            folder_slides_folders = [i['folderId'] for i in all_folder_slides]
            unique_folders = list(set(folder_slides_folders))
            folders_in_folder = []
            for u in unique_folders:
                if not u==folder_info['_id'] and not u in folders_in_folder:
                    # This is for all folders in this folder
                    # This grabs parent folders of this folder
                    u_folder_info = self.handler.get_folder_info(folder_id=u, user_token = user_external_token)
                    u_folder_rootpath = self.handler.get_folder_rootpath(u, user_token = user_external_token)
                    # Folders in order from collection-->child folder-->etc.
                    folder_ids = [i['object']['_id'] for i in u_folder_rootpath]

                    if any([i['object']['_modelType']=='collection' for i in u_folder_rootpath]):
                        base_model = 'collection'
                    else:
                        base_model = 'user'

                    if folder_ids[-1]==folder_info['_id']:
                        child_folder_path = f'/{base_model}/'+'/'.join([i['object']['name'] if i['object']['_modelType'] in ['folder','collection'] else i['object']['login'] for i in u_folder_rootpath]+[u_folder_info['name']])
                    else:
                        # Folder that is immediate child of current folder:
                        child_folder_idx = folder_ids.index(folder_info['_id'])
                        child_folder_path = f'/{base_model}/'+'/'.join([i['object']['name'] if i['object']['_modelType'] in ['folder','collection'] else i['object']['login'] for i in u_folder_rootpath[:child_folder_idx+2]])


                    child_folder_path_info = self.handler.get_path_info(
                        path = child_folder_path,
                        user_token = user_external_token
                    )
                    if not child_folder_path_info['_id'] in folders_in_folder:
                        folders_in_folder.append(child_folder_path_info['_id'])
                        
                        # Adding folder to list if the number of items is above zero or show_empty is True
                        folder_folders.append({
                            'Name': child_folder_path_info['name'],
                            '_id': child_folder_path_info['_id'],
                            'Type': child_folder_path_info['_modelType'],
                            'Number of Folders': child_folder_path_info['nFolders'],
                            'Number of Slides': child_folder_path_info['nItems'],
                            'Last Updated': child_folder_path_info['updated']
                        } | child_folder_path_info['meta'])

                elif u==folder_info['_id']:
                    # This means that there are some slides that are direct children (not in a sub-folder) in this folder. 
                    # This adds them all at once
                    for i in all_folder_slides:
                        if i['folderId']==folder_info['_id']:
                            folder_slides.append(
                                {
                                    'Name': i['name'],
                                    '_id': i['_id'],
                                    'Type': i['_modelType'],
                                    'Last Updated':i['updated']
                                } | {k:v for k,v in i['meta'].items() if type(v)==str}
                            )

            if show_empty:
                # This is how you get all the empty folders within a folder (does not get child empty folders)
                empty_folders = self.handler.get_folder_folders(
                    folder_id = folder_info['_id'],
                    folder_type = folder_info['_modelType'],
                    user_token = user_external_token
                )
                
                for f in empty_folders:
                    if not f['_id'] in folders_in_folder and not f['_id'] in unique_folders:
                        folder_info = self.handler.gc.get(f'/folder/{f["_id"]}/details?token={user_external_token}')
                        folder_folders.append(
                            {
                                'Name': f['name'],
                                '_id': f['_id'],
                                'Type': f['_modelType'],
                                'Number of Folders': folder_info['nFolders'],
                                'Number of Slides': folder_info['nItems'],
                                'Last Updated': f['updated']
                            }
                        )

        elif folder_info['_modelType']=='user':
            
            user_folders = ['Private','Public']
            for u_f in user_folders:
                user_folder_info = self.handler.get_path_info(
                    path = f'/user/{folder_info["login"]}/{u_f}',
                    user_token = user_external_token
                )

                folder_folders.append({
                    'Name': user_folder_info['name'],
                    '_id': user_folder_info['_id'],
                    'Type': user_folder_info['_modelType'],
                    'Number of Folders': user_folder_info['nFolders'],
                    'Number of Slides': user_folder_info['nItems'],
                    'Last Updated': user_folder_info['updated']
                })

        elif folder_info['_modelType']=='item':
            
            if self.selector_type=='file':
                item_subs = self.handler.gc.get(f'/item/{folder_info["_id"]}/files?token={user_external_token}')
            elif self.selector_type=='annotation':
                item_subs = self.handler.gc.get(f'/annotation?token={user_external_token}',parameters={'itemId': folder_info['_id']})

            folder_slides = [
                {
                    'Name': i['name'] if self.selector_type=='file' else i['annotation']['name'],
                    '_id': i['_id'],
                    'Type': i['_modelType'],
                }
                for i in item_subs
            ]


        return folder_slides, folder_folders

    def get_callbacks(self):

        # Updating resource list based on selection (from table, from crumbs), return internal folders/slides
        self.blueprint.callback(
            [
                Input({'type': 'dsa-resource-selector-resource-table','index': ALL},'selected_rows'),
                Input({'type': 'dsa-resource-selector-path-part','index': ALL},'n_clicks'),
                Input({'type':'dsa-resource-selector-output-name-button','index': ALL},'n_clicks')
            ],
            [
                State({'type': 'dsa-resource-selector-resource-table','index': ALL},'data'),
                State({'type': 'dsa-resource-selector-current-resource-path','index': ALL},'children'),
                State({'type': 'dsa-resource-selector-output-name-input','index': ALL},'value'),
                State({'type': 'anchor-dsa-resource-selector-selected-resources','index': ALL},'data'),
                State('anchor-vis-store','data')
            ],
            [
                Output({'type': 'dsa-resource-selector-resource-table-div','index': ALL},'children'),
                Output({'type': 'dsa-resource-selector-current-resource-path','index':ALL},'children'),
                Output({'type': 'anchor-dsa-resource-selector-selected-resources','index': ALL},'data')
            ],
            prevent_initial_call = True
        )(self.update_resource_selection)

    def update_resource_selection(self, table_selected_rows, table_path_clicked, ouput_file_name_clicked, table_data, current_path_parts, output_file_name, selected_resource_data, session_data):

        current_path_parts = list(self.extract_path_parts(get_pattern_matching_value(current_path_parts)))
        
        output_file_name = get_pattern_matching_value(output_file_name)

        session_data = json.loads(session_data)

        user_external_login = self.get_user_external_login(session_data)
        user_external_token = self.get_user_external_token(session_data)

        selected_resource_data = json.loads(get_pattern_matching_value(selected_resource_data))
        added_resources = []
        removed_resources = []
        current_resources = [
            i['_id'] for i in selected_resource_data['resource_list']
        ]

        if 'dsa-resource-selector-resource-table' in ctx.triggered_id['type']:
            # This is so that it works for multiple selected rows as well
            table_selected = ctx.triggered_id['index']
            selected_resource = [table_data[table_selected][i] for i in table_selected_rows[table_selected]]
            resource_type = [i['Type'] for i in selected_resource]
            if not len(selected_resource)==0:
                if not all([r==self.selector_type for r  in resource_type]):
                    # This means another table should be added:
                    if not selected_resource[0]['Name']=='User Folders':
                        resource_slides, resource_folders = self.organize_folder_contents(
                            folder_info = {
                                '_id': selected_resource[0]['_id'],
                                '_modelType': selected_resource[0]['Type']
                            },
                            show_empty = True,
                            ignore_histoqc=True,
                            session_data=session_data
                        )
                    else:
                        resource_slides, resource_folders = self.organize_folder_contents(
                            folder_info = {
                                '_modelType': 'user',
                                'login': user_external_login,
                                'token': user_external_token
                            },
                            show_empty=True,
                            ignore_histoqc=True,
                            session_data=session_data
                        )

                    if not any([i in current_path_parts for i in ['user','collection']]):
                        if selected_resource[0]['Type']=='collection':
                            current_path_parts += ['collection']
                        elif selected_resource[0]['Type']=='user':
                            current_path_parts += ['user']

                    if not selected_resource[0]['Name']=='User Folders':
                        new_crumbs = current_path_parts + [selected_resource[0]['Name']]
                    else:
                        new_crumbs = current_path_parts
                else:
                    resource_folders = None
                    resource_slides = None

                    added_resources = [
                        i for i in selected_resource if not i['_id'] in current_resources
                    ]
                    removed_resources = [
                        i for i in table_data[table_selected] if i['_id'] in current_resources and not i in selected_resource
                    ]
            else:
                removed_resources = selected_resource_data['resource_list']
                resource_folders = None
                resource_slides = None

        elif 'dsa-resource-selector-path-part' in ctx.triggered_id['type']:
            selected_resource = current_path_parts[ctx.triggered_id['index']]
            if ctx.triggered_id['index']<len(current_path_parts)-1:
                if ctx.triggered_id['index']>1:
                    new_path = '/'.join(current_path_parts[1:ctx.triggered_id['index']+1])
                    path_info = self.handler.get_path_info(new_path,user_token = user_external_token)

                    resource_slides, resource_folders = self.organize_folder_contents(
                        folder_info = path_info,
                        show_empty = True,
                        ignore_histoqc=True,
                        session_data=session_data
                    )

                    new_crumbs = current_path_parts[:ctx.triggered_id['index']+1]

                elif ctx.triggered_id['index']==1:
                    if selected_resource=='collection':
                        resource_folders = self.get_collection_table(session_data,return_data=True)
                        resource_slides = []
                        new_crumbs = current_path_parts[:ctx.triggered_id['index']+1]
                    elif selected_resource=='user':
                        resource_slides, resource_folders = self.organize_folder_contents(
                            folder_info = {
                                '_modelType': 'user',
                                'login': user_external_login
                            },
                            show_empty = True,
                            ignore_histoqc=True,
                            session_data=session_data
                        )
                        new_crumbs = current_path_parts[:ctx.triggered_id['index']+1]

                elif ctx.triggered_id['index']==0:
                    resource_folders = self.get_collection_table(session_data, return_data=True)                   
                    resource_slides = []
                    new_crumbs = [current_path_parts[0]]

                else:
                    resource_folders = None
                    resource_slides = None

            else:
                new_crumbs = current_path_parts
                resource_folders = None
                resource_slides = None

        elif 'dsa-resource-selector-output-name-button' in ctx.triggered_id['type']:
            
            if not 'user' in current_path_parts:
                new_path = '/'.join(current_path_parts[1:])
            else:
                new_path = f'/user/{session_data["user"]["login"]}/'+'/'.join(current_path_parts[2:])
            path_info = self.handler.get_path_info(new_path,user_token = user_external_token)

            if not output_file_name is None:
                added_resources = [
                    {
                        'folderId': path_info['_id'],
                        'fileName': output_file_name,
                        'Name': output_file_name
                    }
                ]

            resource_folders = None
            resource_slides = None

        if not all([i is None for i in [resource_folders,resource_slides]]):
            if all([type(i)==list for i in [resource_folders,resource_slides]]):
                resource_folder_table = self.make_selectable_dash_table(
                    dataframe=pd.DataFrame.from_records(resource_folders),
                    id = {'type': f'dsa-resource-selector-resource-table','index': 0},
                    selected_rows = [idx for idx,i in enumerate(resource_folders) if i['_id'] in current_resources],
                    use_prefix = True
                )

                resource_slides_table = self.make_selectable_dash_table(
                    dataframe=pd.DataFrame.from_records(resource_slides),
                    id = {'type': f'{self.component_prefix}-dsa-resource-selector-resource-table','index': 1},
                    selected_rows=[idx for idx,i in enumerate(resource_slides) if i['_id'] in current_resources],
                    use_prefix=True
                )

                if not self.selector_channel=='output':
                    updated_resource_table = [
                        dbc.Stack([
                            resource_folder_table,
                            html.Hr(),
                            html.H5('Slides in selected resource: '),
                            resource_slides_table
                        ])
                    ]
                else:
                    # Check that this isn't a user base folder
                    if not current_path_parts==['Collections and User Folders','user',user_external_login]:
                        input_component = html.Div([
                            dbc.InputGroup([
                                dbc.InputGroupText(
                                    'Output File Name: '
                                ),
                                dbc.Input(
                                    id = {'type': 'dsa-resource-selector-output-name-input','index': 0},
                                    placeholder = 'Output File Name',
                                    type = 'text',
                                    required = True,
                                    value = [],
                                    maxLength = 1000
                                ),
                                dbc.Button(
                                    'Done!',
                                    color = 'success',
                                    n_clicks = 0,
                                    id = {'type': 'dsa-resource-selector-output-name-button','index':0}
                                )
                            ])
                        ],style = {'marginTop':'10px'})

                        PrefixIdTransform(prefix = f'{self.component_prefix}',escape=lambda input_id: self.prefix_escape(input_id)).transform_layout(input_component)

                        updated_resource_table = [
                            dbc.Stack([
                                resource_folder_table,
                                html.Hr(),
                                input_component,
                                html.Hr(),
                                html.H5('Slides in selected resource: '),
                                resource_slides_table
                            ])
                        ]

            else:
                updated_resource_table = [resource_folders]

            updated_resource_path = [
                dbc.Stack([
                    html.A(
                        c+'/',
                        n_clicks = 0,
                        id = {'type': f'dsa-resource-selector-path-part','index': c_idx},
                        style = {'color': 'rgb(0,0,255)'}
                    )
                    for c_idx,c in enumerate(new_crumbs)
                ],direction='horizontal')
            ]

            PrefixIdTransform(prefix=f'{self.component_prefix}',escape = lambda input_id: self.prefix_escape(input_id)).transform_layout(updated_resource_path[0])

        else:
            updated_resource_table = [no_update]
            updated_resource_path = [no_update]

        if len(added_resources)==0 and len(removed_resources)==0:
            updated_resource_data = [no_update]
        else:
            # Updating current selected resources
            for d in removed_resources:
                selected_resource_data['resource_list'].remove(d)
            new_resource_list = {
                'resource_list': selected_resource_data['resource_list']+added_resources,
                'source_component_index': selected_resource_data['source_component_index']
            }
            updated_resource_data = [json.dumps(new_resource_list)]

        return updated_resource_table, updated_resource_path, updated_resource_data




import os
import pandas as pd
import numpy as np
import json
import uuid

from datetime import datetime

# Dash imports
import dash
dash._dash_renderer._set_react_version('18.2.0')
from dash import dcc, ALL, MATCH, ctx, exceptions, no_update
import dash_bootstrap_components as dbc
import dash_mantine_components as dmc
from dash_extensions.enrich import DashProxy, html, MultiplexerTransform, PrefixIdTransform, Input, State, Output

from typing_extensions import Union
from fusion_tools.tileserver import Slide, TileServer, DSATileServer, LocalTileServer, CustomTileServer
from fusion_tools.database.database import fusionDB
from fusion_tools.database.api import fusionAPI

import threading

import uvicorn
from fastapi import FastAPI
from fastapi.middleware.wsgi import WSGIMiddleware
from fastapi.middleware.cors import CORSMiddleware
import asyncio
import nest_asyncio

from pathlib import Path
import dash_uploader as du

class Visualization:
    """General holder class used for initialization. Components added after initialization.
    To initialize a new visualization session, you can use the following syntax:

    .. code-block:: python

        # This is for a slide stored on the same computer you're running the fusion-tools instance from.
        local_slide = ['/path/to/slide.tif']
        annotations = ['/path/to/annotations.json']
        components = [
            [
                SlideMap()
            ],
            [
                [
                    OverlayOptions(),
                    PropertyViewer()
                ]
            ]
        ]
        vis_session = Visualization(
            local_slides = local_slide,
            local_annotations = local_annotations,
            components = components
        )
        vis_session.start()
        
    """
    
    def __init__(self,
                 local_slides: Union[Slide,list,str,None] = None,
                 local_annotations: Union[list,dict,None] = None,
                 slide_metadata: Union[list,dict,None] = None,
                 tileservers: Union[list,TileServer,None] = None,
                 database: Union[fusionDB, None] = None,
                 components: Union[list,dict] = [],
                 header: list = [],
                 app_options: dict = {},
                 linkage: Union[list,str] = 'row'
                 ):
        """Constructor method

        :param local_slides: Filepath for individual slide or list of filepaths stored locally, defaults to None
        :type local_slides: Union[list,str,None], optional
        :param local_annotations: List of processed annotations or filepaths for annotations stored locally (aligns with local_slides), defaults to None
        :type local_annotations: Union[list,dict,None], optional
        :param slide_metadata: List or single dictionary containing slide-level metadata keys and values, defaults to None
        :type slide_metadata: Union[list,dict,None], optional
        :param tileservers: Single tileserver or multiple tileservers, defaults to None
        :type tileservers: Union[list,TileServer,None], optional
        :param components: List of components in layout format (rows-->columns-->tabs for nested lists) or dictionary containing names of pages and their component lists, defaults to []
        :type components: Union[list,dict], optional
        :param header: List of components to add to collapsed upper portion of the visualization, defaults to None
        :type header: Union[list,None], optional
        :param app_options: Additional options for the running visualization session, defaults to {}
        :type app_options: dict, optional
        :param linkage: Which levels of components are linked through callbacks (can be 'row','col',or 'tab'), defaults to 'row'
        :type linkage: str, optional
        """

        self.local_slides = local_slides
        self.slide_metadata = slide_metadata
        self.tileservers = tileservers
        self.database = database
        self.local_annotations = local_annotations
        self.components = components
        self.header = header
        self.app_options = app_options
        self.linkage = linkage

        self.access_count = 0

        # New parameter defining how unique components can be linked
        # page = components in the same page can communicate
        # row = components in the same row can communicate
        # col = components in the same column can communicate
        # tab = components in the same tab can communicate
        if type(self.linkage)==str:
            assert self.linkage in ['page','row','col','tab']
        elif type(self.linkage)==list:
            # Only applies for multi-page layouts
            assert type(self.components)==dict
            assert all([i in ['page','row','col','tab'] for i in self.linkage])

        self.default_options = {
            'title': 'FUSION',
            'assets_folder': os.path.join(os.getcwd(),'.fusion_assets')+os.sep,
            'requests_pathname_prefix':'/',
            'server': 'default',
            'server_options': {},
            'port': 8080,
            'jupyter': False,
            'host': 'localhost',
            'debug': False,
            'layout_style': {},
            'cors': {
                'allow_origins': ['*'],
                'allow_methods': ['GET','OPTIONS'],
                'allow_credentials': False,
                'allow_headers': ['*'],
                'expose_headers': ['*'],
                'max_age': '36000000'
            },
            'external_stylesheets': [
                dbc.themes.LUX,
                dbc.themes.BOOTSTRAP,
                dbc.icons.BOOTSTRAP,
                dbc.icons.FONT_AWESOME,
                dmc.styles.ALL
            ],
            'transforms': [
                MultiplexerTransform()
            ],
            'external_scripts': [
                'https://cdnjs.cloudflare.com/ajax/libs/chroma-js/2.1.0/chroma.min.js',
            ]
        }

        # Where default options are merged with user-added options
        self.app_options = self.default_options | self.app_options

        self.default_page = self.app_options.get('default_page')

        self.assets_folder = self.app_options['assets_folder']
        if not os.path.exists(self.assets_folder):
            os.makedirs(self.assets_folder)

        self.initialize_database()

        self.vis_store_content = self.initialize_stores()

        self.viewer_app = DashProxy(
            __name__,
            url_base_pathname = None if not self.app_options['jupyter'] else self.app_options.get('url_base_pathname',None),
            requests_pathname_prefix = self.app_options.get('requests_pathname_prefix','/'),
            suppress_callback_exceptions = True,
            external_stylesheets = self.app_options['external_stylesheets'],
            external_scripts = self.app_options['external_scripts'],
            assets_folder = self.assets_folder,
            prevent_initial_callbacks=True,
            transforms = self.app_options['transforms']
        )

        self.viewer_app.title = self.app_options['title']
        self.viewer_app.layout = self.gen_layout()
        self.get_callbacks()

    def initialize_database(self):

        #TODO: Find some way to make it easier for components to connect to this database
        if self.database is None:
            print(f'Creating fusionDB instance at: {self.app_options.get("assets_folder","")}fusion_database.db')
            if os.path.exists(self.app_options.get("assets_folder","")+'fusion_database.db'):
                print(f'Removing previous instance of fusionDB')
                print(self.app_options.get('assets_folder','')+'fusion_database.db')
                os.unlink(self.app_options.get('assets_folder','')+'fusion_database.db')

            self.database = fusionDB(
                db_url = f'sqlite:///{self.app_options.get("assets_folder","")}fusion_database.db',
                echo = False
            )

        elif type(self.database)==str:
            print(f'Creating fusionDB instance at: {self.database}')
            self.database = fusionDB(
                db_url = self.database,
                echo = False
            )

    def get_callbacks(self):

        self.viewer_app.callback(
            [
                Input('anchor-page-url','pathname'),
                Input('anchor-page-url','search'),
                Input({'type':'page-button','index': ALL},'n_clicks')
            ],
            [
                State('anchor-vis-store','data'),
                State('anchor-vis-memory-store','data')
            ],
            [
                Output('vis-container','children'),
                Output('user-current','children'),
                Output('anchor-vis-store','data'),
                Output('anchor-vis-memory-store','data')
            ],
            prevent_initial_call = True
        )(self.update_page)

        self.viewer_app.callback(
            [
                Input('navbar-toggler','n_clicks')
            ],
            [
                Output('navbar-collapse','is_open')
            ],
            [
                State('navbar-collapse','is_open')
            ]
        )(self.open_navbar)

        self.viewer_app.callback(
            [
                Input('header-toggler','n_clicks')
            ],
            [
                Output('header-collapse','is_open')
            ],
            [
                State('header-collapse','is_open')
            ]
        )(self.open_header)

        self.viewer_app.callback(
            [
                Input({'type': 'header-button','index': ALL},'n_clicks')
            ],
            [
                Output('header-modal','is_open'),
                Output('header-modal','size'),
                Output('header-modal','fullscreen'),
                Output('header-modal','className'),
                Output('header-modal','children')
            ],
            [
                State('anchor-vis-store','data')
            ]
        )(self.open_header_component)

        self.viewer_app.callback(
            [
                Input('user-current','n_clicks')
            ],
            [
                Output('login-modal','is_open'),
                Output('login-modal','children')
            ],
            [
                State('anchor-vis-store','data'),
                State('anchor-vis-memory-store','data')
            ]
        )(self.open_current_modal)

        # Callback(s) for logging in
        self.viewer_app.callback(
            [
                Input('user-login','n_clicks')
            ],
            [
                Output('login-modal','children')
            ],
            [
                State('anchor-vis-store','data'),
                State('anchor-vis-memory-store','data')
            ]
        )(self.open_login_modal)

        # Callback for creating a new user
        self.viewer_app.callback(
            [
                Input({'type': 'user-login-new', 'index': ALL},'n_clicks')
            ],
            [
                Output('login-modal','children')
            ],
            [
                State('anchor-vis-store','data'),
                State('anchor-vis-memory-store','data')
            ]
        )(self.open_new_user_modal)

        # Callback for submitting login/ submitting create new user
        self.viewer_app.callback(
            [
                Input({'type': 'user-login-new-submit','index': ALL},'n_clicks'),
                Input({'type': 'user-login-submit','index': ALL},'n_clicks'),
                Input({'type': 'user-logout-submit','index': ALL},'n_clicks'),
                Input({'type': 'user-info-save','index': ALL},'n_clicks')
            ],
            [
                Output('anchor-vis-memory-store','data'),
                Output('anchor-vis-store','data'),
                Output('anchor-page-url','pathname'),
                Output('login-modal','is_open')
            ],
            [
                State('anchor-vis-store','data'),
                State('anchor-vis-memory-store','data'),
                State('anchor-page-url','pathname'),
                State({'type': 'user-prev-key','index': ALL},'children'),
                State({'type': 'user-prev-val','index': ALL},'value'),
                State({'type': 'user-new-key','index': ALL},'children'),
                State({'type': 'user-new-val','index': ALL},'value'),
                State({'type': 'user-current-key','index': ALL},'children'),
                State({'type': 'user-current-val','index': ALL},'value')
            ]
        )(self.update_current_user)

    def open_navbar(self, clicked, is_open):

        if clicked:
            return not is_open
        return is_open
    
    def open_header(self, clicked, is_open):

        if clicked:
            return not is_open
        return is_open

    def open_header_component(self, clicked,session_data):

        if clicked:
            session_data = json.loads(session_data)
            header_open = True
            header_children = self.header[ctx.triggered_id['index']].update_layout(session_data = session_data, use_prefix = True)
            if hasattr(self.header[ctx.triggered_id['index']],'modal_size'):
                header_size = self.header[ctx.triggered_id['index']].modal_size
            else:
                header_size = 'lg'
            
            if hasattr(self.header[ctx.triggered_id['index']],'fullscreen'):
                header_fullscreen = self.header[ctx.triggered_id['index']].fullscreen
            else:
                header_fullscreen = False

            if hasattr(self.header[ctx.triggered_id['index']],'modal_className'):
                header_className = self.header[ctx.triggered_id['index']].modal_className
            else:
                header_className = None


            return header_open, header_size, header_fullscreen, header_className, header_children
        else:
            raise exceptions.PreventUpdate

    def open_current_modal(self, clicked, session_data, in_memory_store):
        
        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate

        session_data = json.loads(session_data)
        in_memory_store = json.loads(in_memory_store)
        user_info = in_memory_store.get('user')
        session_info = session_data.get('session')
        user_FirstLast = f'{user_info.get("firstName")} {user_info.get("lastName")}'

        if 'guest' in user_info.get('id'):
            user_type = 'guest'
            user_badge = dmc.Badge(
                "guest", variant='filled'
            )
        elif user_info.get('admin'):
            user_type = 'admin'
            user_badge = dmc.Badge(
                'admin',variant='filled',color='red'
            )
        else:
            user_type = 'user'
            user_badge = dmc.Badge(
                'user', variant='filled',color='green'
            )

        user_avatar = dbc.Row([
                dbc.Col(
                    [
                        dmc.Avatar(id='user-avatar',name=user_FirstLast,color='initial',size='lg',radius='xl'),
                        dbc.Tooltip(user_info.get('login'),target='user-avatar',placement='top')
                    ],
                    md = 'auto'
                ),
                dbc.Col(
                    html.H5(f'{user_FirstLast}'),
                    md = 'auto'
                ),
                dbc.Col(
                    user_badge,
                    md = 'auto'
                )
            ],align='center'
        )

        # Note- Clipboard components don't show up in Chrome
        # Guest users can't change info, User/Admin users can
        if user_type in ['user','admin']:
            user_info_rows = []
            for idx, (key,value) in enumerate(user_info.items()):
                if key in ['id','token']:
                    input_row = html.P(
                        value,
                        id= {'type': 'user-current-val','index': idx}
                    )
                elif key in ['login','firstName','lastName','email']:
                    input_row = dbc.Input(
                        value=value,
                        type='text',
                        id={'type': 'user-current-val','index': idx}
                    )

                elif key=='external':
                    input_row = dmc.Badge(
                        children = [user_info.get('external',{}).get('login')],
                        color = 'teal',
                        variant = 'dot',
                        size = 'xl',
                        id = {'type': 'user-current-val','index': idx}
                    )

                else:
                    # Skipping password, admin, updated
                    continue

                user_info_rows.append(
                    dbc.Row([
                        dbc.Label(
                            key,
                            id = {'type': 'user-current-key','index': idx},
                            html_for={'type': 'user-current-val','index': idx},
                            width=2
                        ),
                        dbc.Col([
                            input_row
                        ],md=9),
                        dbc.Col([
                            dcc.Clipboard(
                                target_id = {'type': 'user-current-val','index': idx},
                                title = 'copy',
                                style = {
                                    'fontSize':20,
                                    'color': 'rgb(200,200,200)'
                                }
                            )
                        ],md=1)
                    ],align = 'center')
                )

            button_rows = dbc.Row([
                dbc.Col(
                    dbc.Button(
                        "Save Changes",
                        id = {'type': 'user-info-save','index': 0},
                        n_clicks = 0,
                        color = 'secondary',
                        className = 'd-grid col-12 mx-auto'
                    ),
                    md = 'auto'
                ),
                dbc.Col(
                    dbc.Button(
                        "Log Out",
                        id = {'type': 'user-logout-submit','index': 0},
                        n_clicks = 0,
                        color = 'danger',
                        className = 'd-grid col-12 mx-auto',
                        disabled = False
                    ),
                    md = 'auto'
                ),
            ])

        else:
            user_info_rows = []
            for idx, (key,value) in enumerate(user_info.items()):
                if not value is None:
                    if not key=='external':
                        user_info_rows.append(
                            dbc.Row([
                                dbc.Label(
                                    key,
                                    id = {'type': 'user-current-key','index': idx},
                                    html_for={'type': 'user-current-val','index': idx},
                                    width=2
                                ),
                                dbc.Col([
                                    html.P(
                                        value,
                                        id={'type': 'user-current-val','index': idx}
                                    )
                                ],md=9),
                                dbc.Col([
                                    dcc.Clipboard(
                                        target_id = {'type': 'user-current-val','index': idx},
                                        title = 'copy',
                                        style = {
                                            'fontSize':20,
                                            'color': 'rgb(200,200,200)'
                                        }
                                    )
                                ],md=1)
                            ],align = 'center')
                       )
                    elif key=='external':
                        user_info_rows.append(
                            dbc.Row([
                                dbc.Label(
                                    key,
                                    width = 2
                                ),
                                dbc.Col([
                                    dmc.Badge(
                                        children = [user_info.get('external',{}).get('login')],
                                        color = 'teal',
                                        variant = 'dot',
                                        size = 'xl',
                                        id = 'user-external'
                                    ),
                                ],md = 10)
                            ])
                        )
            

            button_rows = dbc.Row([
                dbc.Col(
                    dbc.Button(
                        "Login",
                        id = 'user-login',
                        n_clicks = 0,
                        color = 'success',
                        className = 'd-grid col-12 mx-auto'
                    ),
                    md = 12
                )
            ])

        #TODO: For User/Admin users, show a dropdown menu of session ids with name (id)
        session_info_rows = dbc.Row([
            dbc.Label(
                'session',
                html_for = 'user-session-row',
                width=2
            ),
            dbc.Col([
                html.P(
                    session_info.get('id'),
                    id = 'user-session-row'
                )
            ])
        ])

        modal_children = [
            dbc.ModalHeader([
                dbc.ModalTitle("Current User Actions")
            ]),
            dbc.ModalBody([
                user_avatar,   
                html.Hr(),
            ] + user_info_rows + [html.Hr(),session_info_rows]+[
                html.Hr(),
                button_rows
            ]),
            dbc.ModalFooter([
                dbc.Row([
                    dbc.Col([
                        html.P(
                            'fusion-tools v.3.6.35',
                            style={'font-style':'italics'}
                        )
                    ],md='auto'),
                    dbc.Col([
                        html.A(
                            'API',
                            href='/docs',
                            target='_blank'
                        )
                    ],md='auto')
                ]), 
            ])
        ]

        if clicked:
            return True, modal_children
        else:
            raise exceptions.PreventUpdate

    def open_login_modal(self, clicked, session_data, in_memory_store):

        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate
        
        modal_children = [
            dbc.ModalHeader([
                dbc.ModalTitle("Log in")
            ]),
            dbc.ModalBody([
                dbc.Row([
                    dbc.Label(
                        'login',
                        id = {'type': 'user-prev-key','index': 0},
                        html_for = {'type': 'user-prev-val','index': 0},
                        width = 2
                    ),
                    dbc.Col(
                        dbc.Input(
                            type = 'text', 
                            id = {'type': 'user-prev-val','index': 0}, 
                            placeholder = 'login'
                        ),
                        md = 10
                    )
                ],align='center'),
                dbc.Row([
                    dbc.Label(
                        'password',
                        id = {'type': 'user-prev-key','index': 1},
                        html_for = {'type': 'user-prev-val','index': 1},
                        width = 2
                    ),
                    dbc.Col(
                        dbc.Input(
                            type = 'password', 
                            id = {'type': 'user-prev-val','index': 1}, 
                            placeholder = 'password'
                        )
                    )
                ],align='center'),
                html.Hr(),
                dbc.Row([
                    dbc.Col([
                            dbc.Button(
                            'Submit',
                            color = 'success',
                            id = {'type':'user-login-submit','index': 0},
                            n_clicks = 0,
                            className = 'd-grid col-12 mx-auto'
                        )
                    ],md=6),
                    dbc.Col([
                        dbc.Button(
                            'Create New User',
                            color = 'primary',
                            id = {'type': 'user-login-new','index': 0},
                            n_clicks = 0,
                            className = 'd-grid col-12 mx-auto'
                        )
                    ],md=6)
                ])
            ]),
            dbc.ModalFooter([
                dbc.Row([
                    dbc.Col([
                        html.P(
                            'fusion-tools v.3.6.35',
                            style={'font-style':'italics'}
                        )
                    ],md='auto'),
                    dbc.Col([
                        html.A(
                            'API',
                            href='/docs',
                            target='_blank'
                        )
                    ],md='auto')
                ]), 
            ])
        ]
        
        if clicked:
            return modal_children
        else:
            raise exceptions.PreventUpdate

    def open_new_user_modal(self, clicked, session_data, in_memory_store):

        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate
        
        modal_children = [
            dbc.ModalHeader([
                dbc.ModalTitle("Create New User")
            ]),
            dbc.ModalBody([
                dbc.Row([
                    dbc.Label(
                        'firstName',
                        id = {'type': 'user-new-key','index': 0},
                        html_for = {'type': 'user-new-val','index': 0},
                        width=2
                    ),
                    dbc.Col([
                        dbc.Input(
                            type = 'text',
                            id = {'type': 'user-new-val','index': 0},
                            placeholder = 'First Name'
                        )
                    ],md=10)
                ],align='center'),
                dbc.Row([
                    dbc.Label(
                        'lastName',
                        id = {'type': 'user-new-key','index': 1},
                        html_for = {'type': 'user-new-val','index': 1},
                        width=2
                    ),
                    dbc.Col([
                        dbc.Input(
                            type = 'text',
                            id = {'type': 'user-new-val','index': 1},
                            placeholder = 'Last Name'
                        )
                    ],md=10)
                ],align='center'),
                dbc.Row([
                    dbc.Label(
                        'email',
                        id = {'type': 'user-new-key','index': 2},
                        html_for = {'type': 'user-new-val','index': 2},
                        width = 2
                    ),
                    dbc.Col([
                        dbc.Input(
                            type = 'email',
                            id = {'type': 'user-new-val','index': 2},
                            placeholder = 'example@email.com (OPTIONAL)'
                        )
                    ])
                ]),
                dbc.Row([
                    dbc.Label(
                        'login',
                        id = {'type': 'user-new-key','index': 3},
                        html_for = {'type': 'user-new-val','index': 3},
                        width = 2
                    ),
                    dbc.Col(
                        dbc.Input(
                            type = 'text', 
                            id = {'type': 'user-new-val','index': 3}, 
                            placeholder = 'login'
                        ),
                        md = 10
                    )
                ],align='center'),
                dbc.Row([
                    dbc.Label(
                        'password',
                        id = {'type': 'user-new-key','index': 4},
                        html_for = {'type': 'user-new-val','index': 4},
                        width = 2
                    ),
                    dbc.Col(
                        dbc.Input(
                            type = 'password', 
                            id = {'type': 'user-new-val','index': 4}, 
                            placeholder = 'password'
                        )
                    )
                ],align='center'),
                html.Hr(),
                dbc.Row([
                    dbc.Col([
                        dbc.Button(
                            'Create New User',
                            color = 'primary',
                            id = {'type': 'user-login-new-submit','index': 0},
                            n_clicks = 0,
                            className = 'd-grid col-12 mx-auto'
                        )
                    ],md=12)
                ])
            ]),
            dbc.ModalFooter([
                dbc.Row([
                    dbc.Col([
                        html.P(
                            'fusion-tools v.3.6.35',
                            style={'font-style':'italics'}
                        )
                    ],md='auto'),
                    dbc.Col([
                        html.A(
                            'API',
                            href='/docs',
                            target='_blank'
                        )
                    ],md='auto')
                ]), 
            ])
        ]
        
        if clicked:
            return modal_children
        else:
            raise exceptions.PreventUpdate

    def update_current_user(self, new_clicked, login_clicked, logout_clicked, update_info_clicked, session_data, in_memory_store, pathname, prev_keys, prev_vals, new_keys, new_vals, current_keys, current_vals):

        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate

        session_data = json.loads(session_data)
        in_memory_store = json.loads(in_memory_store)

        if 'user-login-new-submit' in ctx.triggered_id['type']:
            # Check new user info is present
            new_user_info = {
                k:v
                for k,v in zip(new_keys,new_vals)
                if not k=='external'
            } | {
                'admin': False,
                'token': uuid.uuid4().hex[:24],
                'external': in_memory_store.get('external')
            }

            new_user = self.database.create_new_user(
                user_kwargs = new_user_info
            )

            if new_user is None:
                # Some failure creating new user
                raise exceptions.PreventUpdate
            
            else:
                new_in_memory = {
                    'user': new_user,
                    'session': in_memory_store.get('session')
                }

        elif 'user-login-submit' in ctx.triggered_id['type']:
            # Check returning user info
            returning_user_info = {
                k:v
                for k,v in zip(prev_keys,prev_vals)
                if not k=='external'
            } | {'external': in_memory_store.get('external')}

            previous_user = self.database.check_user_login_password(
                prev_vals[0],prev_vals[1]
            )

            if previous_user is None:
                # Password failed/login not present
                raise exceptions.PreventUpdate
            else:
                new_in_memory = {
                    'user': previous_user,
                    'session': in_memory_store.get('session')
                }

        elif 'user-logout-submit' in ctx.triggered_id['type']:
            
            new_in_memory = {
                'user': None,
                'session': None
            }

        elif 'user-info-save' in ctx.triggered_id['type']:
            
            current_keys.append('token')
            current_vals.append(uuid.uuid4().hex[:24])

            updated_user = self.database.get_create(
                table_name = 'user',
                inst_id = in_memory_store.get('user').get('id'),
                kwargs = {k:v for k,v in zip(current_keys,current_vals) if not k in ['id','external']} | {'external': in_memory_store.get('user').get('external')}
            )
            if not updated_user is None:
                new_in_memory = {
                    'user': updated_user.to_dict(),
                    'session': in_memory_store.get('session')
                }
            else:
                raise exceptions.PreventUpdate

        new_session_data = session_data
        new_session_data['user'] = new_in_memory.get('user')
        new_session_data['session'] = new_in_memory.get('session')

        return json.dumps(new_in_memory,default=str), json.dumps(new_session_data,default=str), pathname+'#', False

    def update_page(self, pathname, path_search, path_button, session_data, in_memory_store):
        """Updating page in multi-page application

        :param pathname: Pathname or suffix of current url which is a key to the page name
        :type pathname: str
        """
        session_data = json.loads(session_data)
        in_memory_store = json.loads(in_memory_store)

        pathname = pathname.replace('#','')

        self.access_count +=1
        # session_data is preserved in the tab (not cleared on refresh)
        # in_memory_store is preserved in that instance of that tab (cleared on refreshing)
        current_user_ids = self.database.get_ids('user')
        current_vis_session_ids = self.database.get_ids('vis_session')

        if self.access_count == 1:
            print('-----------------First Access-------------')
            # This is the first time the app has been accessed, set to the created guest User and VisSession
            in_memory_store['user'] = self.database.get_user(
                user_id = self.database.get_ids('user')[0]
            )
            del in_memory_store['user']['updated']

            in_memory_store['session'] = {
                'id': self.database.get_ids('vis_session')[0]
            }

            session_data['user'] = in_memory_store['user']
            session_data['session'] = in_memory_store['session']

            session_data['current'] = self.vis_store_content.get('current')
            session_data['local'] = self.vis_store_content.get('local')

        elif self.access_count > 1:
            
            current_user_ids = self.database.get_ids('user')
            current_vis_session_ids = self.database.get_ids('vis_session')

            if in_memory_store.get('user') is None:
                if session_data.get('user') is None:
                    print(f'---------New Window/New User/New Session-------------')
                    # This is a new user, hasn't entered the application from this tab
                    new_user = self.new_user(guest = True)
                    new_session = self.new_session(guest = True)

                    session_data['current'] = [i for i in self.vis_store_content.get('current') if i.get('public')]
                    session_data['local'] = self.vis_store_content.get('local')

                    in_memory_store['user'] = new_user
                    in_memory_store['session'] = new_session
                    session_data['user'] = new_user
                    session_data['session'] = new_session

                    # Adding vis_session
                    self.database.add_vis_session(
                        session_data
                    )

                elif session_data.get('user') is not None:
                    if session_data.get('user').get('id') in current_user_ids:
                        if session_data.get('session').get('id') in current_vis_session_ids:
                            print(f'---------New Window/Previous User/Previous Session-------------')
                            in_memory_store['user'] = session_data.get('user')
                            in_memory_store['session'] = session_data.get('session')

                            # Checking which items this user has specific access to
                            prev_user_access = self.database.check_user_access(user_id = in_memory_store.get('user').get('id'), admin = in_memory_store.get('user').get('admin',False))
                            # Adding public items and items this user has access to
                            session_data['current'] = [i for i in self.vis_store_content.get('current') if i.get('public') or i.get('id') in prev_user_access]
                            session_data['local'] = self.vis_store_content.get('local')

                        else:
                            print(f'---------New Window/Previous User/New Session-------------')
                            # Creating a new session id
                            new_session = self.new_session(guest = not 'guest' in session_data.get('user').get('id'))
                            # Checking which items this user has specific access to
                            prev_user_access = self.database.check_user_access(user_id = session_data.get('user').get('id'), admin = session_data.get('user').get('admin',False))
                            # Adding public items and items this user has access to
                            session_data['current'] = [i for i in self.vis_store_content.get('current') if i.get('public') or i.get('id') in prev_user_access]
                            session_data['local'] = self.vis_store_content.get('local')

                            in_memory_store['session'] = new_session
                            session_data['session'] = new_session
                            in_memory_store['user'] = session_data.get('user')
                            in_memory_store['session'] = session_data.get('session')

                            # Adding vis_session
                            self.database.add_vis_session(
                                session_data
                            )
                    else:
                        print(f'---------New Window/New User/New Session-------------')
                        # This is a new user, hasn't entered the application from this tab
                        new_user = self.new_user(guest = True)
                        new_session = self.new_session(guest = True)

                        session_data['current'] = [i for i in self.vis_store_content.get('current') if i.get('public')]
                        session_data['local'] = self.vis_store_content.get('local')

                        in_memory_store['user'] = new_user
                        in_memory_store['session'] = new_session
                        session_data['user'] = new_user
                        session_data['session'] = new_session

                        # Adding vis_session
                        self.database.add_vis_session(
                            session_data
                        )

            elif in_memory_store.get('user').get('id') not in current_user_ids:
                print('--------Previous Window/New User/New Session---------------')
                # Not None user, not in current_user_ids
                new_user = self.new_user(guest = False, id = in_memory_store.get('user').get('id'))
                new_session = self.new_session(guest = not 'guest' in in_memory_store.get('user').get('id'))

                # Since this user is not registered in User, they only have access to 'public' Items
                session_data['current'] = [i for i in self.vis_store_content.get('current') if i.get('public')]
                session_data['local'] = self.vis_store_content.get('local')

                in_memory_store['user'] = new_user
                in_memory_store['session'] = new_session
                session_data['user'] = new_user
                session_data['session'] = new_session
            
            elif in_memory_store.get('user').get('id') in current_user_ids:
                if in_memory_store.get('session').get('id') in current_vis_session_ids:
                    print('------------Previous Window/Previous User/Previous Session---------------')
                    # Checking which items this user has specific access to
                    prev_user_access = self.database.check_user_access(user_id = in_memory_store.get('user').get('id'), admin = in_memory_store.get('user').get('admin',False))
                    # Adding public items and items this user has access to
                    session_data['current'] = [i for i in session_data.get('current') if i.get('public') or i.get('id') in prev_user_access]
                    session_data['local'] = session_data.get('local')

                else:
                    print('------------Previous Window/Previous User/New Session---------------')
                    # Not None user, in current_user_ids
                    # Creating a new session id
                    new_session = self.new_session(guest = not 'guest' in in_memory_store.get('user'))
                    # Checking which items this user has specific access to
                    prev_user_access = self.database.check_user_access(user_id = in_memory_store.get('user').get('id'), admin = in_memory_store.get('user').get('admin',False))
                    # Adding public items and items this user has access to
                    session_data['current'] = [i for i in self.vis_store_content.get('current') if i.get('public') or i.get('id') in prev_user_access]
                    session_data['local'] = self.vis_store_content.get('local')

                    in_memory_store['session'] = new_session
                    session_data['user'] = in_memory_store.get('user')
                    session_data['session'] = new_session

                    # Adding vis_session
                    self.database.add_vis_session(
                        session_data
                    )

        user_name = f'{in_memory_store.get("user").get("firstName")} {in_memory_store.get("user").get("lastName","")}'
        signed_in_user = f'Signed in as: {user_name}'

        page_content = no_update
        page_pathname = no_update
        page_search = no_update

        if ctx.triggered_id=='anchor-page-url':
            if pathname in self.layout_dict:
                # If that path is in the layout dict, return that page content

                # If the page needs to be updated based on changes in anchor-vis-data
                page_content = self.update_page_layout(
                    page_components_list = self.components[pathname.replace(self.app_options.get('requests_pathname_prefix','/'),'').replace('-',' ')],
                    use_prefix = True,
                    session_data=session_data
                )
            
            elif 'session' in pathname:
                from fusion_tools.handler.dsa_handler import DSAHandler
                temp_handler = DSAHandler(
                    girderApiUrl=os.environ.get('DSA_URL')
                )
                session_content = temp_handler.get_session_data(path_search.replace('?id=',''))
                new_session_data = {
                    'current': session_content['current'],
                    'local': session_data['local'],
                    'data': session_content['data'],
                }

                if 'user' in session_data:
                    new_session_data['user'] = session_data['user']
                
                page_pathname = session_content['page'].replace(self.app_options.get('requests_pathname_prefix','/'),'').replace('-',' ')

                page_content = self.update_page_layout(
                    page_components_list = self.components[page_pathname],
                    use_prefix=True,
                    session_data=new_session_data
                )

                in_memory_store['session'] = {
                    'id': path_search.replace('?id=','')
                }
            
            elif 'item' in pathname:
                #TODO: Loading an individual item from id

                pass

            else:
                if self.default_page is None:
                    # Otherwise, return a list of clickable links for valid pages
                    page_content = html.Div([
                        html.H1('Uh oh!'),
                        html.H2(f'The page: {pathname}, is not in the current layout!'),
                        html.Hr()
                    ] + [
                        html.P(html.A(page,href=page))
                        for page in self.layout_dict
                    ])

                else:
                    page_content = self.update_page_layout(
                        page_components_list = self.components[self.default_page.replace(self.app_options.get('requests_pathname_prefix','/'),'').replace('-',' ')],
                        use_prefix = True,
                        session_data=session_data
                    )

        elif ctx.triggered_id['type']=='page-button':
            new_pathname = list(self.layout_dict.keys())[ctx.triggered_id['index']]
            # If the page needs to be updated based on changes in anchor-vis-data
            page_content = self.update_page_layout(
                page_components_list = self.components[new_pathname.replace(self.app_options.get('requests_pathname_prefix','/'),'').replace('-',' ')],
                use_prefix = True,
                session_data=session_data
            )
    
        return page_content, signed_in_user, json.dumps(session_data), json.dumps(in_memory_store)

    def new_user(self, guest: bool = True, id: Union[str,None] = None):

        user_dict = {
            'id': f'guestuser{uuid.uuid4().hex[:15]}' if guest and id is None else id,
            'login': f'{uuid.uuid4().hex[:24]}',
            'firstName': 'Guest',
            'lastName': 'User',
            'token': uuid.uuid4().hex[:24]
        }

        return user_dict

    def new_session(self, guest: bool = True):

        session_dict = {
            'id': f'guestsession{uuid.uuid4().hex[:12]}' if guest else uuid.uuid4().hex[:24],
            'data': {}
        }

        return session_dict

    def initialize_stores(self):

        # This should be all the information necessary to reproduce the tileservers and annotations for each image
        slide_store = {
            "current": [],
            "local": [],
            "data": {},
            'session': self.new_session(guest = True),
            'user': self.new_user(guest = True),
        }

        self.database.add_vis_session(slide_store)

        s_idx = 0
        t_idx = 0

        if not self.local_slides is None:
            if self.local_annotations is None:
                self.local_annotations = [None]*len(self.local_slides)
            
            if self.slide_metadata is None:
                self.slide_metadata = [None]*len(self.local_slides)

            # Initializing a LocalTileServer instance.
            self.local_tile_server = LocalTileServer(
                tile_server_port = self.app_options['port'],
                host = self.app_options['host'],
                database = self.database,
                jupyter_server_url = self.app_options.get('jupyter_server_url','').replace(str(self.app_options['port']),str(self.app_options['port']+10))
            )

            for s_idx,(s,anns,meta) in enumerate(zip(self.local_slides,self.local_annotations,self.slide_metadata)):
                slide_dict = {}
                if not s is None:
                    # Adding this slide to list of local slides
                    local_slide_id = uuid.uuid4().hex[:24]
                    if not isinstance(s,Slide):
                        self.local_tile_server.add_new_image(
                            new_image_id = local_slide_id,
                            new_image_path = s,
                            new_annotations = anns,
                            new_metadata = meta,
                            session_id = slide_store.get('session').get('id'),
                            user_id = slide_store.get('user').get('id')
                        )

                        slide_dict = {
                            'name': s.split(os.sep)[-1],
                            'id': local_slide_id,
                            'cached': True,
                            'public': False,
                            'type': 'local_item',
                        } | self.local_tile_server.get_slide_urls(local_slide_id)
                    else:
                        self.local_tile_server.add_new_slide(
                            slide_id = local_slide_id,
                            slide_obj = s,
                            session_id = slide_store.get('session').get('id'),
                            user_id = slide_store.get('user').get('id') if not s.public else None
                        )

                        slide_dict = {
                            'name': s.image_filepath.split(os.sep)[-1],
                            'id': local_slide_id,
                            'cached': True,
                            'public': s.public,
                            'type': 'local_item',
                        } | self.local_tile_server.get_slide_urls(local_slide_id)

                slide_store['current'].append(slide_dict)
                slide_store['local'].append(slide_dict)

        else:
            self.local_tile_server = None

        if not self.tileservers is None:
            if isinstance(self.tileservers,TileServer):
                self.tileservers = [self.tileservers]
            
            for t_idx,t in enumerate(self.tileservers):
                if type(t)==LocalTileServer:
                    local_tile_server_url = f'{t.protocol}://{t.host}:{t.port}/tileserver'

                    slide_store['current'].extend([
                        {
                            'name': j.name,
                            'id': j.id,
                            'url': local_tile_server_url,
                            'cached': True,
                            'type': 'local_item',
                        } | t.get_slide_urls(j.id, standalone = True)
                        for j_idx,j in enumerate(t.get_item_names_ids())
                    ])
                    slide_store['local'].extend([
                        {
                            'name': j.name,
                            'id': j.id,
                            'url': local_tile_server_url,
                            'cached': True,
                            'type': 'local_item',
                        } | j.get_slide_urls(j.id,standalone=True)
                        for j_idx,j in enumerate(t.get_item_names_ids())
                    ])

                elif type(t)==DSATileServer:

                    #TODO: Update "public" as needed, it's possible this will have some other access control.
                    slide_store['current'].append({
                        'name': t.name,
                        'id': t.item_id,
                        'remote_id': t.item_id,
                        'url': t.base_url,
                        'public': True,
                        'type': 'remote_item',
                    })
                elif type(t)==CustomTileServer:
                    slide_store['current'].append({
                        'name': t.name,
                        'id': t.id,
                    })

        return slide_store

    def gen_layout(self):
        """Generating Visualization layout

        :return: Total layout containing embedded components
        :rtype: dmc.MantineProvider
        """
        self.get_layout_children()

        title_nav_bar = dbc.Navbar(
            dbc.Container([
                dcc.Location(id='anchor-page-url',refresh=False),
                dbc.Row([
                    dbc.Col([
                        html.Div([
                            html.H3(self.app_options['title'],style={'color': 'rgb(255,255,255)'})
                        ])
                    ],md=True,align='center')
                ],align='center'),
                dbc.Row([
                    dbc.Col([
                        dbc.NavbarToggler(id='navbar-toggler'),
                        dbc.Collapse(
                            dbc.Nav([
                                dbc.NavItem(
                                    dbc.DropdownMenu(
                                        children = [
                                            dbc.DropdownMenuItem(
                                                dbc.Button(
                                                    page,
                                                    id = {'type': 'page-button','index':page_idx},
                                                    n_clicks = 0,
                                                    className = 'd-grid col-12 mx-auto',
                                                    style = {'textTransform': 'none'}
                                                )
                                            )
                                            for page_idx,page in enumerate(self.layout_dict)
                                        ] + [
                                            dbc.DropdownMenuItem(divider=True),
                                            dbc.DropdownMenuItem([
                                                dbc.Button(
                                                    'Signed in as: Guest User',
                                                    id = 'user-current',
                                                    color = 'secondary',
                                                    outline = True,
                                                    n_clicks = 0,
                                                    style = {
                                                        'font-style': 'italic',
                                                        'textTransform': 'none'
                                                    }
                                                ),
                                                dbc.Modal(
                                                    id = 'login-modal',
                                                    centered = True,
                                                    is_open = False,
                                                    size = 'xl',
                                                    className = None,
                                                    children = []
                                                )
                                            ])
                                        ],
                                        label = 'Menu',
                                        nav = True,
                                        style = {
                                            'border-radius':'5px'
                                        }
                                    )
                                ),
                                dbc.NavItem(
                                    dbc.Button(
                                        'Examples',
                                        id = 'examples-button',
                                        outline=True,
                                        color='primary',
                                        href='https://spborder.github.io/fusion-welcome-page/',
                                        target='_blank',
                                        style={'textTransform':'none'}
                                    )
                                ),
                                dbc.NavItem(
                                    dbc.Button(
                                        'Docs',
                                        id = 'docs-button',
                                        outline=True,
                                        color='primary',
                                        href='https://fusion-tools.readthedocs.io/en/latest/',
                                        target='_blank',
                                        style={'textTransform':'none'}
                                    )
                                ),
                                dbc.NavItem(
                                    dbc.Button(
                                        'Issues',
                                        id = 'issues-button',
                                        outline=True,
                                        color='primary',
                                        href='https://github.com/spborder/fusion-tools/issues',
                                        target='_blank',
                                        style={'textTransform':'none'}
                                    )
                                )
                            ],navbar=True),
                            id = 'navbar-collapse',
                            navbar=True
                        )
                    ],md=2)
                ],align='center')
            ],fluid=True),
            dark=True,
            color='dark',
            sticky='fixed',
            style={'marginBottom':'20px'}
        )

        vis_data = html.Div([
            dcc.Store(
                id = 'anchor-vis-store',
                data = json.dumps({}),
                storage_type = 'session',
                modified_timestamp = -1
            ),
            dcc.Store(
                id = 'anchor-vis-memory-store',
                data = json.dumps({}),
                storage_type='memory',
                modified_timestamp = -1
            )
        ])

        self.app_start_time = datetime.now()
        print(f'app start time: {self.app_start_time}')

        if len(self.header)>0:
            header_components = self.gen_header_components()
        else:
            header_components = html.Div()

        layout = dmc.MantineProvider(
            children = [
                vis_data,
                html.Div(
                    dbc.Container(
                        id = 'full-vis-container',
                        fluid = True,
                        children = [title_nav_bar] + [header_components]+ [html.Div(id = 'vis-container',children = [])]
                    ),
                    style = self.app_options['app_style'] if 'app_style' in self.app_options else {}
                )
            ]
        )

        return layout

    def update_page_layout(self, page_components_list:list, use_prefix:bool, session_data:Union[list,dict]):
        
        page_children = []
        for row_idx,row in enumerate(page_components_list):
            row_children = []
            if type(row)==list:
                for col_idx, col in enumerate(row):
                    if not type(col)==list:
                        # If this component needs to be updated with new session data, call that method here
                        if col.session_update:
                            col_layout = col.update_layout(
                                session_data = session_data, 
                                use_prefix = use_prefix
                            )
                        else:
                            # If it doesn't need to be updated, get the layout as is
                            col_layout = col.blueprint.layout

                        row_children.append(
                            dbc.Col(
                                dbc.Card([
                                    dbc.CardHeader(
                                        col.title
                                    ),
                                    dbc.CardBody(
                                        col_layout
                                    )
                                ]),
                                width = True
                            )
                        )

                    else:
                        tabs_children = []
                        for tab_idx, tab in enumerate(col):
                            if tab.session_update:
                                tab_layout = tab.update_layout(
                                    session_data = session_data,
                                    use_prefix = use_prefix
                                )
                            else:
                                tab_layout = tab.blueprint.layout

                            tabs_children.append(
                                dbc.Tab(
                                    dbc.Card(
                                        dbc.CardBody(
                                            tab_layout
                                        )
                                    ),
                                    label = tab.title,
                                    tab_id = tab.title.lower().replace(' ','-')
                                )
                            )

                        row_children.append(
                            dbc.Col(
                                dbc.Card([
                                    dbc.CardHeader('Tools'),
                                    dbc.CardBody(
                                        dbc.Tabs(
                                            tabs_children,
                                            id = {'type': f'{col[0].component_prefix}-vis-layout-tabs','index': np.random.randint(0,1000)},
                                            active_tab = col[0].title.lower().replace(' ','-')
                                        )
                                    )
                                ]),
                                width = True
                            )
                        )

            else:
                
                if row.session_update:
                    row_layout = row.update_layout(
                        session_data = session_data,
                        use_prefix = use_prefix
                    )
                else:
                    row_layout = row.blueprint.layout

                row_children.append(
                    dbc.Col(
                        dbc.Card([
                            dbc.CardHeader(row.title),
                            dbc.CardBody(
                                row_layout
                            )
                        ]),
                        width = True
                    )
                )
            
            page_children.append(
                dbc.Row(
                    row_children
                )
            )

        return page_children

    def get_layout_children(self):
        """Generating layout of embedded components from structure of components list

        :return: List of dbc.Row(dbc.Col(dbc.Tabs())) components
        :rtype: list
        """

        n_rows = 1
        n_cols = 1
        n_tabs = 0

        component_prefix = 0
        self.layout_dict = {}
        page_components = []
        row_components = []
        col_components = []
        tab_components = []

        if type(self.components)==list:
            n_rows = len(self.components)
            if any([type(i)==list for i in self.components]):
                n_cols = max([len(i) for i in self.components if type(i)==list])

                if any([any([type(j)==list for j in i]) for i in self.components if type(i)==list]):
                    n_tabs = max([max([len(i) for i in j if type(i)==list]) for j in self.components if type(j)==list])

            print(f'------Creating Visualization with {n_rows} rows, {n_cols} columns, and {n_tabs} tabs--------')
            print(f'----------------- Components in the same {self.linkage} may communicate through callbacks---------')
        
            self.components = {
                'main': self.components 
            }

            self.default_page = 'main'

        elif type(self.components)==dict:
            for page in self.components:
                n_cols = 1
                n_tabs = 0
                n_rows = len(self.components[page])
                if any([type(i)==list for i in self.components[page]]):
                    n_cols = max([len(i) for i in self.components[page] if type(i)==list])

                    if any([any([type(j)==list for j in i]) for i in self.components[page] if type(i)==list]):
                        n_tabs = max([max([len(i) for i in j if type(i)==list]) for j in self.components[page] if type(j)==list])

                print(f'------Creating Visualization Page {page} with {n_rows} rows, {n_cols} columns, and {n_tabs} tabs--------')
                print(f'----------------- Components in the same {self.linkage} may communicate through callbacks---------')
            

        # Iterating through each named page
        for page_idx,page in enumerate(list(self.components.keys())):
            page_children = []
            
            if type(self.linkage)==str:
                if self.linkage=='page':
                    component_prefix = page_idx
            elif type(self.linkage)==list:
                if self.linkage[page_idx]=='page':
                    component_prefix = page_idx

            row_components = []
            for row_idx,row in enumerate(self.components[page]):
                
                if type(self.linkage)==str:
                    if self.linkage=='row':
                        component_prefix = row_idx
                elif type(self.linkage)==list:
                    if self.linkage[page_idx]=='row':
                        component_prefix = row_idx

                row_children = []
                if type(row)==list:
                    col_components = []
                    for col_idx,col in enumerate(row):

                        if type(self.linkage)==str:
                            if self.linkage=='col':
                                component_prefix = col_idx
                        elif type(self.linkage)==list:
                            if self.linkage[page_idx]=='col':
                                component_prefix = col_idx

                        if not type(col)==list:
                            col.add_assets_folder(self.assets_folder)
                            col.load(component_prefix = component_prefix)
                            col.gen_layout(session_data = self.vis_store_content)
                            col.add_database(database = self.database)
                            col_components.append(str(col))
                            
                            row_children.append(
                                dbc.Col(
                                    dbc.Card([
                                        dbc.CardHeader(
                                            col.title
                                        ),
                                        dbc.CardBody(
                                            col.blueprint.embed(self.viewer_app)
                                        )
                                    ]),
                                    width = True
                                )
                            )
                        else:
                            tab_components = []
                            tabs_children = []
                            for tab_idx,tab in enumerate(col):

                                if type(self.linkage)==str:
                                    if self.linkage=='tab':
                                        component_prefix = tab_idx
                                elif type(self.linkage)==list:
                                    if self.linkage[page_idx]=='tab':
                                        component_prefix = tab_idx
                                
                                tab.add_assets_folder(self.assets_folder)
                                tab.load(component_prefix = component_prefix)
                                tab.gen_layout(session_data = self.vis_store_content)
                                tab.add_database(database = self.database)
                                tab_components.append(str(tab))

                                tabs_children.append(
                                    dbc.Tab(
                                        dbc.Card(
                                            dbc.CardBody(
                                                tab.blueprint.embed(self.viewer_app)
                                            )
                                        ),
                                        label = tab.title,
                                        tab_id = tab.title.lower().replace(' ','-')
                                    )
                                )
                            col_components.append(tab_components)

                            row_children.append(
                                dbc.Col(
                                    dbc.Card([
                                        dbc.CardHeader('Tools'),
                                        dbc.CardBody(
                                            dbc.Tabs(
                                                tabs_children,
                                                id = {'type': f'vis-layout-tabs','index': np.random.randint(0,1000)},
                                                active_tab=col[0].title.lower().replace(' ','-')
                                            )
                                        )
                                    ]),
                                    width = True
                                )
                            )
                
                    row_components.append(col_components)
                else:
                    
                    row.add_assets_folder(self.assets_folder)
                    row.load(component_prefix = component_prefix)
                    row.gen_layout(session_data = self.vis_store_content)
                    row.add_database(database = self.database)
                    row_components.append(str(row))

                    row_children.append(
                        dbc.Col(
                            dbc.Card([
                                dbc.CardHeader(row.title),
                                dbc.CardBody(
                                    row.blueprint.embed(self.viewer_app)
                                )
                            ]),
                            width = True
                        )
                    )

                page_children.append(
                    dbc.Row(
                        row_children
                    )
                )

            page_components.append(row_components)
            self.layout_dict[self.app_options.get('requests_pathname_prefix','/')+page.replace(" ","-")] = page_children

        upload_check = self.check_for_uploader(page_components)
        if upload_check:
            from fusion_tools.handler.dataset_uploader import DSAUploadHandler
            du.configure_upload(
                self.viewer_app, 
                Path(self.assets_folder+'/tmp/uploads'),
                upload_api = '/upload',
                http_request_handler = DSAUploadHandler
            )

    def check_for_uploader(self, components):
        check = False
        for c in components:
            if type(c)==list:
                check = check | self.check_for_uploader(c)
            elif type(c)==str:
                if c=='DSA Uploader':
                    check = True
        
        return check

    def gen_header_components(self):
        
        for h_idx, h in enumerate(self.header):
            h.load(h_idx)
            h.gen_layout({})
            h.add_database(database = self.database)

        header_components = html.Div([
            html.Hr(),
            dbc.Modal(
                id = 'header-modal',
                centered = True,
                is_open = False,
                size = 'xl',
                className = None,
                children = [
                    h.blueprint.embed(self.viewer_app)
                    for h_idx,h in enumerate(self.header)
                ]
            ),
            dbc.Row([
                dbc.Col([
                    dbc.NavbarToggler(id = 'header-toggler'),
                    dbc.Collapse(
                        dbc.Nav([
                            dbc.NavItem(
                                dbc.Button(
                                    h.title,
                                    id = {'type': 'header-button','index': h_idx},
                                    outline = True,
                                    color = 'primary',
                                    style = {'textTransform':'none'}
                                )
                            )
                            for h_idx,h in enumerate(self.header)
                        ],navbar=False),
                        id = 'header-collapse',
                        navbar=False,
                        is_open = True
                    )
                ])
            ]),
            html.Hr()
        ])

        return header_components

    def create_app(self):

        app = FastAPI(
            title = 'FUSION',
            description = 'Modular visualization and analysis dashboard creation for high-resolution microscopy images.',
            version='3.6.42',
            docs_url='/docs',
            redoc_url = '/redoc',
            license_info = {
                'name': 'Apache 2.0'
            }
        )

        app.include_router(
            fusionAPI(
                database = self.database
            ).router
        )

        allowed_origins = [
            'http://localhost',
            f'http://localhost:{self.app_options.get("port")}',
            f'http://0.0.0.0:{self.app_options.get("port")}',
            'http://0.0.0.0',
            f'http://127.0.0.1:{self.app_options.get("port")}',
            'http://127.0.0.1',
            'http://'+self.app_options.get('host'),
            'http://'+self.app_options.get('host')+':'+str(self.app_options.get('port'))
        ]

        if not self.app_options.get('jupyter_server_url') is None:
            allowed_origins.append(
                self.app_options.get('jupyter_server_url')
            )

        allowed_origins += [
            i.replace('http://','')
            for i in allowed_origins
        ]
        allowed_origins += [
            i+'/'
            for i in allowed_origins
        ]

        allowed_origins = list(set(allowed_origins))
        cors_options = self.app_options.get('cors')

        if self.local_tile_server is not None:
            self.local_tile_server.app.include_router(self.local_tile_server.router)
            app.mount(
                path = self.app_options.get('requests_pathname_prefix','/')+'tileserver/',
                app = CORSMiddleware(
                    self.local_tile_server.app,
                    allow_origins = allowed_origins if not cors_options.get('allow_origins')==['*'] else ['*'],
                    allow_methods = cors_options.get('allow_methods',['GET','OPTIONS']),
                    allow_headers = cors_options.get('allow_headers',['*']),
                    allow_credentials = cors_options.get('allow_credentials',False),
                    expose_headers = cors_options.get('expose_headers',['*']),
                    max_age = cors_options.get('max_age','36000000')
                )
            )
        
        app.mount(
            path = self.app_options.get('requests_pathname_prefix','/'), 
            app = CORSMiddleware(
                WSGIMiddleware(self.viewer_app.server),
                    allow_origins = allowed_origins if not cors_options.get('allow_origins')==['*'] else ['*'],
                    allow_methods = cors_options.get('allow_methods',['GET','OPTIONS']),
                    allow_headers = cors_options.get('allow_headers',['*']),
                    allow_credentials = cors_options.get('allow_credentials',False),
                    expose_headers = cors_options.get('expose_headers',['*']),
                    max_age = cors_options.get('max_age','36000000')
            )
        )

        return CORSMiddleware(
            app,
            allow_origins = allowed_origins if not cors_options.get('allow_origins')==['*'] else ['*'],
            allow_methods = cors_options.get('allow_methods',['GET','OPTIONS']),
            allow_headers = cors_options.get('allow_headers',['*']),
            allow_credentials = cors_options.get('allow_credentials',False),
            expose_headers = cors_options.get('expose_headers',['*']),
            max_age = cors_options.get('max_age','36000000')
        )

    def start(self):
        """Starting visualization session based on provided app_options
        """

        if not self.app_options['jupyter']:

            uvicorn.run(
                self.create_app(),
                host=self.app_options['host'],
                port=self.app_options['port']
            )

        else:
            if not self.local_tile_server is None:      
                nest_asyncio.apply()      
                new_thread = threading.Thread(
                    target = self.local_tile_server.start,
                    daemon=True
                )
                new_thread.start()

                self.viewer_app.run(
                    host = self.app_options['host'],
                    port = self.app_options['port']
                )

            else:
                self.viewer_app.run(
                    host = self.app_options['host'],
                    port = self.app_options['port']
                )





"""

Structure schemas for different items in SQLite database

"""
import json
import uuid
import time
import bcrypt

from datetime import datetime

from sqlalchemy import (
    not_, or_, and_, func, select, insert, event, union_all, create_engine, update, exc,
    Column, String, Boolean,ForeignKey, JSON, text)
from sqlalchemy.orm import (
    declarative_base, sessionmaker, 
    mapped_column, Session, scoped_session,
    with_loader_criteria
)
from sqlalchemy.pool import NullPool

from typing import Generator
from contextlib import contextmanager
from shapely.geometry import box, shape

from typing_extensions import Union

from .models import (
    Base, User, UserAccess, 
    VisSession, Item, Layer, 
    LocalItem, RemoteItem,
    Structure, ImageOverlay, Annotation,
    Data
)


TABLE_NAMES = {
    'user': User,
    'user_access': UserAccess,
    'vis_session': VisSession,
    'item': Item,
    'local_item': LocalItem,
    'remote_item': RemoteItem,
    'layer': Layer,
    'structure': Structure,
    'image_overlay': ImageOverlay,
    'annotation': Annotation,
    'data': Data
}


#TODO: add access check to UserAccess table for any query that mentions data from an item
# This includes Item, Layer, Structure, ImageOverlay, and Annotation elements
# either user_id or user_token can be used to identify a user, though user_id is a column in UserAccess


class fusionDB:
    def __init__(self,
                 db_url:str,
                 echo:bool = False):
        
        self.engine = create_engine(
            db_url,
            connect_args={
                "check_same_thread": False
            },
            echo = echo,
            pool_pre_ping = True,
            pool_recycle=3600
        )

        Base.metadata.create_all(bind = self.engine)
        self.SessionLocal = scoped_session(sessionmaker(bind=self.engine))

        """
        # Not sure how to add this as an automatic filterer, this will prevent non-public Items from being available on "select" statements
        @event.listens_for(self.SessionLocal,"do_orm_execute")
        def _do_orm_execute(orm_execute_state):
            if (
                orm_execute_state.is_select
                and not orm_execute_state.is_column_load
                and not orm_execute_state.is_relationship_load
            ):
                orm_execute_state.statement = orm_execute_state.statement.options(
                    with_loader_criteria(entity_or_base = Item, where_criteria = Item.public == True)
                )
        """

    @contextmanager
    def get_db(self) -> Generator[Session, None, None]:
        # Generator type has types, yield, send, return (in this case it yields a Session, sends None and returns None)
        db: Session = self.SessionLocal()
        try:
            yield db
            db.commit()
        except Exception:
            db.rollback()
            raise
        finally:
            db.close()

    def get_uuid(self):
        return uuid.uuid4().hex[:24]
    
    def add(self, obj, session = None):
        # Using sessionmaker in a context manager 
        # (closes automatically and rolls back in the event of a database error)
        if session is None:
            with self.get_db() as session:
                session.add(obj)
                session.commit()
        else:
            session.add(obj)

        return True

    def remove(self, obj, session = None):
        # Using sessionmaker in a context manager
        # (closes automatically and rolls back in the event of a database error)
        # Testing if the session object also has a .delete method
        if session is None:
            with self.get_db() as session:
                session.delete(obj)
                session.commit()
        else:
            session.delete(obj)
        
        return True
    
    def get_create(self, table_name:str, inst_id:Union[str,None] = None, kwargs:Union[dict,None] = None):
        """Function for updating or adding items to the database based on whether or not that particular instance already exists

        :param table_name: Name of table that the instance belongs/should belong to
        :type table_name: str
        :param inst_id: A unique id assigned to the new/updated instance, defaults to None
        :type inst_id: Union[str,None], optional
        :param kwargs: Additional arguments used in the creation/updation of an instance, defaults to None
        :type kwargs: Union[dict,None], optional
        :return: Returns the newly created instance if the table name exists 
        """
        with self.get_db() as session:
            if table_name in TABLE_NAMES:
                if not inst_id is None:
                    get_create_result = session.query(
                        TABLE_NAMES.get(table_name)
                    ).filter_by(id = inst_id).first()

                    if not get_create_result:
                        #TODO: Add a "created" option to each model

                        # This is if this thing does not exist in the table
                        updated = datetime.now()
                        get_create_result = TABLE_NAMES.get(table_name)(
                            id = inst_id,
                            **kwargs | {'updated': updated}
                        )

                        self.add(get_create_result, session)
                    else:
                        # This is if this thing DOES exist in the table, update
                        updated = datetime.now()
                        get_create_result = TABLE_NAMES.get(table_name)(
                            id = inst_id,
                            **kwargs | {'updated': updated}
                        )
                        update_kwargs = kwargs | {'updated': updated}
                        session.execute(update(TABLE_NAMES.get(table_name)).where(getattr(TABLE_NAMES.get(table_name),'id')==inst_id).values(update_kwargs))
                
                else:
                    new_id = self.get_uuid()
                    updated = datetime.now()
                    get_create_result = TABLE_NAMES.get(table_name)(
                        id = new_id,
                        **kwargs | {'updated': updated}
                    )

                    self.add(get_create_result, session)

                session.commit()

                return get_create_result
            else:
                return None

    def get_remove(self, table_name:str, inst_id:Union[str,None] = None, user_id: Union[str,None] = None, vis_session_id: Union[str,None] = None):
        """The opposite of self.get_create, checks if an item is present in the table and if it is, deletes it.

        :param table_name: Name of table that the instance belongs to
        :type table_name: str
        :param inst_id: A unique id assigned to the instance, defaults to None
        :type inst_id: Union[str,None], optional
        :return: Returns True if the removal was successful/the table name exists and that instance is in the table
        """
        with self.get_db() as session:
            if table_name in TABLE_NAMES:
                if not inst_id is None:
                    get_remove_result = session.query(
                        TABLE_NAMES.get(table_name)
                    ).filter_by(id = inst_id)

                    if not get_remove_result:
                        #print("Instance not present in db")
                        # If this thing isn't in the table, don't do anything
                        pass
                    else:
                        # This is if this thing DOES exist in the table, remove it
                        if not user_id is None:
                            get_remove_result = get_remove_result.filter(
                                getattr(TABLE_NAMES.get(table_name),'user')==user_id
                            )

                        if not vis_session_id is None:
                            get_remove_result = get_remove_result.filter(
                                getattr(TABLE_NAMES.get(table_name),'session')==vis_session_id
                            )

                        if get_remove_result:
                            for r in get_remove_result.all():
                                self.remove(r,session)
                        else:
                            #print('Not found for this user/vis_session')
                            pass
                else:
                    # Don't delete anything if no instance id is provided
                    get_remove_result = None

                session.commit()

                return get_remove_result
            else:
                return None

    def count(self, table_name:str):
        """Get the count of unique instances within this specific table

        :param table_name: Name of table to query (user, vis_session, item, layer, structure, image_overlay, annotation)
        :type table_name: str
        :return: Count of instances
        :rtype: int
        """

        #TODO: Add some filtering capability here
        with self.get_db() as session:
            if table_name in TABLE_NAMES:
                count_result = session.execute(
                    select(func.count(func.distinct(TABLE_NAMES.get(table_name).id)))
                ).one()

                return count_result[0]
            else:
                return 0

    def search_op(self, search_query, table_name:str, column:str, op:dict):
        """Method for applying different types of filters to database queries

        :param search_query: DB query object
        :type search_query: None
        :param table_name: Name of table being searched
        :type table_name: str
        :param column: Name of column being queried
        :type column: str
        :param op: Operation dictionary
        :type op: dict
        :return: Updated DB query object
        :rtype: None
        """
        op_type = list(op.keys())[0]
        op_query = list(op.values())
        if op_type in ['==','!=','>','<','>=','<=','in','!in']:
            # Numeric operations
            if all([type(i) in [int,float] for i in op_query]) and op_type in ['==','!=','>','<','>=','<=']:
                if op_type=='==':
                    search_query = search_query.filter(
                        getattr(TABLE_NAMES.get(table_name),column)==op_query[0]
                    )
                elif op_type=='!=':
                    search_query = search_query.filter(
                        getattr(TABLE_NAMES.get(table_name),column)!=op_query[0]
                    )
                elif op_type=='>':
                    search_query = search_query.filter(
                        getattr(TABLE_NAMES.get(table_name),column) > op_query[0]
                    )
                elif op_type=='>=':
                    search_query = search_query.filter(
                        getattr(TABLE_NAMES.get(table_name),column) >= op_query[0]
                    )
                elif op_type=='<':
                    search_query = search_query.filter(
                        getattr(TABLE_NAMES.get(table_name),column) < op_query[0]
                    )
                elif op_type=='<=':
                    search_query = search_query.filter(
                        getattr(TABLE_NAMES.get(table_name),column) <= op_query[0]
                    )

            # String operations
            elif all([type(i)==str for i in op_query]) and op_type in ['==','!=']:
                if op_type=='==':
                    search_query = search_query.filter(
                        getattr(TABLE_NAMES.get(table_name),column) == op_query[0]
                    )
                elif op_type=='!=':
                    search_query = search_query.filter(
                        getattr(TABLE_NAMES.get(table_name),column) != op_query[0]
                    )
            
            # List operations
            elif all([type(i)==list for i in op_query]) and op_type in ['in','!in']:
                if op_type=='in':
                    search_query = search_query.filter(
                        getattr(TABLE_NAMES.get(table_name),column).in_(op_query[0])
                    )
                elif op_type=='!in':
                    search_query = search_query.filter(
                        not_(getattr(TABLE_NAMES.get(table_name),column).in_(op_query[0]))
                    )

            return search_query

        else:
            return search_query

    async def search(self, search_kwargs:dict, size:Union[int,None] = None, offset = 0, order = None):
        """Search DB

        :param search_kwargs: Dictionary containing "type" (table name) and "filters"
        :type search_kwargs: dict
        :param size: Number of instances to return, if None returns all that match the search, defaults to None
        :type size: int, optional
        :param offset: Offset of instances to return from search, defaults to 0
        :type offset: int, optional
        :param order: Name of column to order by (Not Implemented), defaults to None
        :type order: None, optional
        :return: Results of DB search
        :rtype: list
        """

        if not search_kwargs.get('type','') in TABLE_NAMES:
            return []

        # By default, not validating user access and only searching public items
        public_only_search = True

        #print(f'db search called: {json.dumps(search_kwargs,indent=4)}')
        with self.get_db() as session:
            search_query = session.query(
                TABLE_NAMES.get(search_kwargs.get('type'))
            )
            
            # Applying filters
            search_filters = search_kwargs.get('filters')
            if not search_filters is None:
                table_joins = list(set(list(search_filters.keys())))

                # Determine ahead of time if will need to check user access
                if search_kwargs.get('type')=='item' and 'user' in table_joins:
                    public_only_search = False

                for t in table_joins:
                    if t in TABLE_NAMES:
                        if hasattr(TABLE_NAMES.get(t),'item') and 'user' in table_joins:
                            public_only_search = False

                if search_kwargs.get('type')=='item' and public_only_search:
                    # This covers the case of public-only-searches
                    search_query = search_query.filter(Item.public == True)
                    
                for t in table_joins:
                    if t in TABLE_NAMES:
                        if not t == search_kwargs.get('type'):
                            # Don't join on User, save that for checking public_only_search
                            if t=='user':
                                continue
                            search_query = search_query.join(TABLE_NAMES.get(t))

                        for k,v in search_filters.get(t).items():
                            if type(v)==str:
                                search_query = search_query.filter(
                                    getattr(TABLE_NAMES.get(t),k)==v
                                )
                            elif type(v)==dict:
                                search_query = self.search_op(search_query,t,k,v)

            if not public_only_search:
                # Combining original query with all public items and all items the user has access to
                user_args = search_filters.get('user')
                query_user = self.get_user(
                    user_id = user_args.get('id'),
                    user_token = user_args.get('token')
                )
                if not query_user is None:
                    # Filtering public or has access
                    search_query = search_query.filter(or_(Item.public==True,Item.user_access.any(id = query_user.get('id'))))
                else:
                    # Filtering public
                    search_query = search_query.filter(Item.public==True)

            return_list = []
            for idx,i in enumerate(search_query.all()):
                if not size is None:
                    if len(return_list)==size:
                        break

                if idx>=offset:
                    return_list.append(i.to_dict())

            return return_list
      
    def add_slide(self,
        slide_id:str, 
        slide_name:str, 
        metadata: dict, 
        image_metadata: dict, 
        image_filepath:Union[str,None], 
        annotations_metadata:dict,
        annotations:Union[list,dict,None],
        vis_session_id: Union[str,None] = None,
        user_id: Union[str,None] = None,
        public: bool = False):

        #print(f'{slide_id=}, {public=}')
        new_item = self.get_create(
            table_name = 'local_item',
            inst_id = slide_id,
            kwargs = {
                'name': slide_name,
                'meta': metadata,
                'image_meta': image_metadata,
                'ann_meta': annotations_metadata,
                'filepath': image_filepath,
                'session': vis_session_id,
                'public': public, 
            }
        )

        if not public and user_id is not None:
            self.add_access(slide_id,user_id)

        for ann_idx, ann in enumerate(annotations):
            # Adding layer
            new_layer = self.get_create(
                table_name = 'layer',
                inst_id = ann['properties']['_id'],
                kwargs = {
                    'name': ann['properties']['name'],
                    'item': slide_id,
                }
            )

            if 'image_path' in ann:
                # Adding image overlay layer
                new_image_overlay = self.get_create(
                    table_name = 'image_overlay',
                    inst_id = ann['properties']['_id'],
                    kwargs = {
                        'bounds': ann['image_bounds'],
                        'properties': ann['image_properties'],
                        'image_src': ann['image_path'],
                        'layer': ann['properties']['_id'],
                        'item': slide_id
                    }
                )
            else:
                # Adding Structures in Layer
                for f_idx, f in enumerate(ann['features']):
                    new_structure = self.get_create(
                        table_name = 'structure',
                        inst_id = f['properties'].get('_id',uuid.uuid4().hex[:24]),
                        kwargs = {
                            'geom': f['geometry'],
                            'properties': f['properties'],
                            'layer': ann['properties']['_id'],
                            'item': slide_id,
                        }
                    )

    def add_vis_session(self, vis_session: dict):

        vis_session_kwargs = {
            'user': vis_session.get('user',{}).get('id'),
            'data': vis_session.get('data',{})
        }

        new_vis_session = self.get_create(
            table_name = 'vis_session',
            inst_id = vis_session.get('session',{}).get('id'),
            kwargs = vis_session_kwargs
        )

        new_user = self.get_create(
            table_name = 'user',
            inst_id = vis_session.get('user',{}).get('id'),
            kwargs = {
                k:v
                for k,v in vis_session.get('user',{}).items() 
                if not k=='id'
            }
        )

    def add_access(self, item_id, user_id):

        with self.get_db() as session:
            statement = session.execute(
                insert(UserAccess).values(
                    user_id = user_id,
                    item_id = item_id
                )
            )
        
    def get_user(self, user_token:Union[str,None] = None, user_id: Union[str,None] = None) -> User:

        with self.get_db() as session:
            if not user_token is None:
                search_query = session.execute(
                    select(User).where(User.token==user_token)
                ).first()
            elif not user_id is None:
                search_query = session.execute(
                    select(User).where(User.id==user_id)
                ).first()

            if not search_query is None:
                if len(search_query)>0:
                    return search_query[0].to_dict()
                else:
                    return None
            else:
                return None

    def check_user_login_password(self, user_login: str, user_password: str):

        with self.get_db() as session:
            search_query = session.execute(
                select(User).where(User.login==user_login)
            ).first()

            if not search_query is None:
                if len(search_query)>0:
                    if search_query[0].verify_password(user_password):
                        return search_query[0].to_dict()
                    else:
                        return None
                else:
                    return None
            else:
                return None

    def create_new_user(self, user_kwargs: dict):

        # Required kwargs are login, password, firstname, lastname
        if not all([user_kwargs.get(j) for j in ['login','password','firstName','lastName']]):
            return None
        else:
            # Check uniqueness of login
            with self.get_db() as session:
                search_query = session.execute(
                    select(User).where(User.login==user_kwargs.get('login'))
                ).first()

                if search_query is None:
                    
                    user_kwargs['password'] = bcrypt.hashpw(user_kwargs['password'].encode(),bcrypt.gensalt())
                    user_kwargs['token'] = self.get_uuid()
                    user_kwargs['updated'] = datetime.now()

                    new_id = self.get_uuid()
                    new_user = User(id = new_id, **user_kwargs)
                    user_kwargs['id'] = new_id

                    self.add(new_user, session)
                    session.commit()

                    return user_kwargs
                else:
                    return None

    def check_user_access(self, user_id:str, admin:bool = False) -> list:

        user_access_rows = []
        with self.get_db() as session:
            user_access = session.query(UserAccess)

            for i in user_access.all():
                user_access_rows.append(i)

        # Admins have access to everything
        non_public_access_list = []
        if admin:
            non_public_access_list =  [i[1] for i in user_access_rows]
        else:
            for user,item in user_access_rows:
                if user_id==user:
                    non_public_access_list.append(item)

        return non_public_access_list

    def get_names(self, table_name:str, user_token: Union[str,None] = None, size:Union[int,None]=None, offset = 0):
        
        #TODO: This should get access controlled
        return_names = []
        with self.get_db() as session:

            if table_name in TABLE_NAMES:
                #if not table_name=='item':
                name_query = session.execute(
                    select(getattr(TABLE_NAMES.get(table_name),'name'))
                )
                #else:
                #    name_query = session.query(Item.name)
                #    name_query = name_query.join(UserAccess)

                for a_idx,a in enumerate(name_query.all()):
                    if not size is None:
                        if len(return_names)==size:
                            break
                    if a_idx>=offset:
                        return_names.append(a[0])

            return return_names

    def get_ids(self, table_name: str, user_token: Union[str,None] = None, size:Union[int,None] = None, offset = 0):

        #TODO: This should be access controlled
        return_ids = []
        with self.get_db() as session:

            if table_name in TABLE_NAMES:
                id_query = session.execute(
                    select(getattr(TABLE_NAMES.get(table_name),'id'))
                )

                for a_idx, a in enumerate(id_query.all()):
                    if not size is None:
                        if len(return_ids)==size:
                            break
                    if a_idx>=offset:
                        return_ids.append(a[0])
            
            return return_ids

    async def get_item_annotations(self, item_id:str, user_id:Union[str,None] = None, vis_session_id:Union[str,None] = None)->list:
        """Loading annotations from item database

        :param item_id: String uuid for an item
        :type item_id: str
        :param user_id: String uuid for a user, defaults to None
        :type user_id: Union[str,None], optional
        :param vis_session_id: String uuid for a visualization session, defaults to None
        :type vis_session_id: Union[str,None], optional
        :return: List of GeoJSON-formatted FeatureCollections (and image overlays if present)
        :rtype: list
        """

        item_annotations = []

        item_layers = await self.search(
            search_kwargs = {
                'type': 'layer',
                'filters': {
                    'item': {
                        'id': item_id
                    }
                }
            }
        )

        for l in item_layers:
            layer_name = l.get('name')
            layer_id = l.get('id')

            layer_structures = await self.search(
                search_kwargs={
                    'type': 'structure',
                    'filters': {
                        'layer': {
                            'id': layer_id
                        }
                    }
                }
            )

            if len(layer_structures)>0:
                item_annotations.append(
                    {
                        'type': 'FeatureCollection',
                        'properties': {
                            'name': layer_name,
                            '_id': layer_id
                        },
                        'features': [
                            {
                                'type': 'Feature',
                                'geometry': s.get('geom'),
                                'properties': s.get('properties')
                            }
                            for s in layer_structures
                        ]
                    }
                )
            else:
                # This could be an ImageOverlay layer
                image_overlays = await self.search(
                    search_kwargs = {
                        'type': 'image_overlay',
                        'filters': {
                            'layer': {
                                'id': layer_id
                            }
                        }
                    }
                )

                if len(image_overlays)>0:
                    item_annotations.extend(
                        [
                            i.to_dict() for i in image_overlays
                        ]
                    )

        return item_annotations

    def get_structure_property_keys(self, item_id:Union[str,list,None] = None, layer_id:Union[str,list,None] = None):
        #TODO: Make a more efficient way to get names of properties for each structure
        pass

    def get_structure_property_data(self, item_id:Union[str,list,None] = None, layer_id:Union[str,list,None] = None, structure_id:Union[str,list,None] = None, property_list:Union[str,list] = None):
        """Extracting one or multiple properties from structures given id filters.

        :param item_id: String uuid for one or multiple image items, defaults to None
        :type item_id: Union[str,list,None], optional
        :param layer_id: String uuid for one or multiple layers, defaults to None
        :type layer_id: Union[str,list,None], optional
        :param structure_id: String uuid for one or multiple structures, defaults to None
        :type structure_id: Union[str,list,None], optional
        :param property_list: List of one or more properties to extract from structures, defaults to None
        :type property_list: Union[str,list], optional
        :return: Records-formatted list of dictionaries with each property, along with structure id, bounding box (minx, miny, maxx, maxy), layer id, layer name, item id, and item name
        :rtype: list
        """

        if property_list is None:
            return []
        elif type(property_list)==str:
            property_list = [property_list]

        # Ensuring uniqueness of property names
        property_list = list(set(property_list))

        with self.get_db() as session:
            search_query = session.query(
                *[Structure.properties[p.split(' --> ')] for p in property_list],
                Structure.id,
                Structure.geom,
                Layer.id,
                Layer.name,
                Item.id,
                Item.name
            ).filter(Structure.layer == Layer.id).filter(Layer.item==Item.id)

            if not item_id is None:
                if type(item_id)==list:
                    search_query = search_query.filter(Item.id.in_(item_id))
                elif type(item_id)==str:
                    search_query = search_query.filter(Item.id == item_id)

            if not layer_id is None:
                if type(layer_id)==list:
                    search_query = search_query.filter(Layer.id.in_(layer_id))
                elif type(layer_id)==str:
                    search_query = search_query.filter(Layer.id==layer_id)

            if not structure_id is None:
                if type(structure_id)==list:
                    search_query = search_query.filter(Structure.id.in_(structure_id))
                elif type(structure_id)==str:
                    search_query = search_query.filter(Structure.id==structure_id)
            
            returned_props = property_list + ['structure.id','geometry','layer.id','layer.name','item.id','item.name']
            return_list = []
            for idx,i in enumerate(search_query.all()):
                i_dict = {'_index': idx}
                for prop,prop_name in zip(i,returned_props):
                    if not prop_name=='geometry':
                        i_dict[prop_name] = prop
                    else:
                        i_dict['bbox'] = list(shape(prop).bounds)

                return_list.append(i_dict)

            return return_list

    def get_structures_in_bbox(self, bbox:list, item_id:Union[str,None] = None, layer_id:Union[str,list,None] = None, structure_id:Union[str,list,None] = None):
        """Querying database for structures that intersect with a 

        :param bbox: _description_
        :type bbox: list
        :param item_id: _description_, defaults to None
        :type item_id: Union[str,None], optional
        :param layer_id: _description_, defaults to None
        :type layer_id: Union[str,list,None], optional
        :param structure_id: _description_, defaults to None
        :type structure_id: Union[str,list,None], optional
        :return: _description_
        :rtype: _type_
        """
        
        #TODO: Test the performance of this using shape().intersects() vs. checking min/max ranges for bounding boxes
        with self.get_db() as session:
            start = time.time()
            search_query = session.query(
                Structure.id,
                Structure.geom
            )

            if not item_id is None:
                if type(item_id)==str:
                    search_query = search_query.filter(Item.id == item_id)

            if not layer_id is None:
                if type(layer_id)==list:
                    search_query = search_query.filter(Layer.id.in_(layer_id))
                elif type(layer_id)==str:
                    search_query = search_query.filter(Layer.id==layer_id)

            if not structure_id is None:
                if type(structure_id)==list:
                    search_query = search_query.filter(Structure.id.in_(structure_id))
                elif type(structure_id)==str:
                    search_query = search_query.filter(Structure.id==structure_id)

            # Box should be minx, miny, maxx, maxy
            query_box = box(*bbox)
            return_list = []
            for idx,i in enumerate(search_query.all()):
                structure_id = i[0]
                structure_geom = i[1]
                if shape(structure_geom).intersects(query_box):
                    return_list.append(structure_id)

            return return_list

    def get_structure_generator(self, item_id: Union[str,list,None] = None, layer_id:Union[str,list,None] = None, structure_id: Union[str,list,None] = None):


        with self.get_db() as session:
            search_query = session.query(
                Structure.id,
                Structure.geom,
                Structure.properties
            ).filter(Structure.layer == Layer.id).filter(Layer.item==Item.id)

            if not item_id is None:
                if type(item_id)==list:
                    search_query = search_query.filter(Item.id.in_(item_id))
                elif type(item_id)==str:
                    search_query = search_query.filter(Item.id == item_id)

            if not layer_id is None:
                if type(layer_id)==list:
                    search_query = search_query.filter(Layer.id.in_(layer_id))
                elif type(layer_id)==str:
                    search_query = search_query.filter(Layer.id==layer_id)

            if not structure_id is None:
                if type(structure_id)==list:
                    search_query = search_query.filter(Structure.id.in_(structure_id))
                elif type(structure_id)==str:
                    search_query = search_query.filter(Structure.id==structure_id)
            
            #returned_props = ['structure.id','geometry','structure.properties']
            #return_list = []
            #for idx,i in enumerate(search_query.all()):
            #    i_dict = {'_index': idx}
            #    for prop,prop_name in zip(i,returned_props):
            #        i_dict[prop_name] = prop

            #    return_list.append(i_dict)

            #return return_list
            return search_query.all()




"""
Defining models for fusionDB
"""
import enum
from typing import List
from sqlalchemy import (
    Table, Column, String, 
    Boolean, Integer, ForeignKey, 
    JSON, DateTime, Enum
)
from sqlalchemy.orm import (
    declarative_base, mapped_column, relationship,
    Mapped    
)
import bcrypt

from shapely.geometry import box

#TODO: Current access control rules specify that if an item is not public then everything associated with that item is also not public
# Likewise, if a user has access to an item, they will have access to everything on that item.

Base = declarative_base()

# Which users can access which items
UserAccess = Table(
    'user_access',
    Base.metadata,
    Column('user_id',String,ForeignKey('user.id',ondelete='CASCADE')),
    Column('item_id',String,ForeignKey('item.id',ondelete='CASCADE'))
)

class User(Base):
    __tablename__ = 'user'
    id = mapped_column(String(24), primary_key = True)
    login = Column(String,unique=True)
    password = Column(String(60))

    firstName = Column(String)
    lastName = Column(String)
    email = Column(String)

    admin = Column(Boolean)
    updated = Column(DateTime)

    token = Column(String)

    item_access: Mapped[List["Item"]] = relationship(
        secondary = UserAccess, back_populates="user_access"
    )

    external = Column(JSON)

    meta = Column(JSON)

    def verify_password(self, query_pword) -> bool:
        return bcrypt.checkpw(
            query_pword.encode(), self.password
        )

    def to_dict(self):
        user_dict = {
            'id': self.id,
            'login': self.login,
            'firstName': self.firstName,
            'lastName': self.lastName,
            'email': self.email,
            'meta': self.meta,
            'external': self.external,
            'admin': self.admin,
            'updated': self.updated,
            'token': self.token
        }

        return user_dict
    


class VisSession(Base):
    __tablename__='vis_session'
    id = mapped_column(String(24),primary_key = True)
    # Can multiple users access the same vis session?
    user = mapped_column(ForeignKey("user.id"))

    name = Column(String)

    #TODO: Add "current" section for current items in vis_session

    # Visualization session data stored as JSON
    data = Column(JSON)
    updated = Column(DateTime)

    meta = Column(JSON)

    def to_dict(self):
        vis_dict = {
            'id': self.id,
            'user': self.user,
            'data': self.data,
            'meta': self.meta,
            'updated': self.updated
        }

        return vis_dict

class ItemType(enum.Enum):
    local = 1
    remote = 2


class Item(Base):
    __tablename__='item'
    id = mapped_column(String(24),primary_key = True)
    name = Column(String)
    meta = Column(JSON)
    image_meta = Column(JSON)
    ann_meta = Column(JSON)

    session = mapped_column(ForeignKey("vis_session.id"))
    updated = Column(DateTime)

    public = Column(Boolean)

    user_access: Mapped[List["User"]] = relationship(
        secondary = UserAccess, back_populates="item_access"
    )
    type = Column(Enum(ItemType))

    __mapper_args__ = {
        'polymorphic_identity': 'item',
        'polymorphic_on': 'type'
    }

    def to_dict(self):
        item_dict = {
            'id': self.id,
            'name': self.name,
            'meta': self.meta,
            'image_meta': self.image_meta,
            'ann_meta': self.ann_meta,
            'session': self.session,
            'updated': self.updated,
            'public': self.public
        }

        return item_dict

class LocalItem(Item):
    __tablename__ = 'local_item'
    id = mapped_column(ForeignKey('item.id'),primary_key = True)

    url = Column(String)
    filepath = Column(String)

    __mapper_args__ = {
        'polymorphic_identity': 'local_item'
    }

    def to_dict(self):

        item_dict = super().to_dict() | {'filepath': self.filepath, 'url': self.url, 'type': self.type}

        return item_dict

class RemoteItem(Item):
    __tablename__ = 'remote_item'
    id = mapped_column(ForeignKey('item.id'),primary_key = True)

    url = Column(String)
    remote_id = Column(String(24))

    __mapper_args__ = {
        'polymorphic_identity': 'remote_item'
    }

    def to_dict(self):

        item_dict = super().to_dict() | {'url': self.url, 'remote_id': self.remote_id, 'type': self.type}

        return item_dict

class Layer(Base):
    __tablename__ = 'layer'
    id = mapped_column(String(24),primary_key = True)
    name = Column(String)
    item = mapped_column(ForeignKey("item.id"))

    updated = Column(DateTime)

    meta = Column(JSON)

    def to_dict(self):
        layer_dict = {
            'id': self.id,
            'name': self.name,
            'item': self.item,
            'meta': self.meta,
            'updated': self.updated
        }

        return layer_dict

class Structure(Base):
    __tablename__ = 'structure'
    id = mapped_column(String(24),primary_key = True)
    geom = Column(JSON)

    properties = Column(JSON)

    layer = mapped_column(ForeignKey('layer.id'))
    item = mapped_column(ForeignKey('item.id'))
    updated = Column(DateTime)

    meta = Column(JSON)

    def to_dict(self):
        structure_dict = {
            'id': self.id,
            'geom': self.geom,
            'properties': self.properties,
            'layer': self.layer,
            'meta': self.meta,
            'updated': self.updated
        }

        return structure_dict
    
    def to_geojson(self):
        geojson_dict = {
            'type': 'Feature',
            'geometry': self.geom,
            'properties': self.properties
        }

        return geojson_dict

class ImageOverlay(Base):
    __tablename__ = 'image_overlay'
    id = mapped_column(String(24),primary_key = True)
    bounds = Column(JSON)

    properties = Column(JSON)
    image_src = Column(String)

    layer = mapped_column(ForeignKey('layer.id'))
    item = mapped_column(ForeignKey('item.id'))
    updated = Column(DateTime)

    meta = Column(JSON)

    def to_dict(self):
        img_overlay_dict = {
            'id': self.id,
            'bounds': self.bounds,
            'properties': self.properties,
            'image_src': self.image_src,
            'layer': self.layer,
            'meta': self.meta,
            'updated': self.updated
        }

        return img_overlay_dict
    
    def to_geojson(self):
        img_overlay_geojson = {
            'type': 'Feature',
            'geometry': list(box(*self.bounds).exterior.coords),
            'properties': self.properties
        }

        return img_overlay_geojson

class Annotation(Base):
    __tablename__ = 'annotation'
    id = Column(String(24),primary_key = True)

    user = mapped_column(ForeignKey('user.id'))
    session = mapped_column(ForeignKey('vis_session.id'))
    item = mapped_column(ForeignKey('item.id'))
    layer = mapped_column(ForeignKey('layer.id'))
    structure = mapped_column(ForeignKey('structure.id'))

    # Storing all annotation data as JSON
    data = Column(JSON)
    updated = Column(DateTime)

    meta = Column(JSON)

    def to_dict(self):
        ann_dict = {
            'id': self.id,
            'user': self.user,
            'session': self.session,
            'item': self.item,
            'layer': self.layer,
            'structure': self.structure,
            'data': self.data,
            'meta': self.meta,
            'updated': self.updated
        }

        return ann_dict

class Data(Base):
    __tablename__ = 'data'
    id = Column(String(24),primary_key = True)

    user = mapped_column(ForeignKey('user.id'))
    session = mapped_column(ForeignKey('vis_session.id'))
    item = mapped_column(ForeignKey('item.id'))
    layer = mapped_column(ForeignKey('layer.id'))
    structure = mapped_column(ForeignKey('structure.id'))

    # filepath to query-able filetype
    filepath = Column(String)
    updated = Column(DateTime)

    meta = Column(JSON)

    def to_dict(self):
        data_dict = {
            'id': self.id,
            'user': self.user,
            'session': self.session,
            'item': self.item,
            'layer': self.item,
            'structure': self.structure,
            'filepath': self.filepath,
            'meta': self.meta,
            'updated': self.updated
        }

        return data_dict

# LocalData & RemoteData?
# Same deal as with LocalItem & RemoteItem




from fusion_tools.components.maps import (
    MapComponent, SlideMap, MultiFrameSlideMap, HybridSlideMap,
    SlideImageOverlay, ChannelMixer, LargeSlideMap, LargeMultiFrameSlideMap
)
from fusion_tools.components.plotters import (
    PropertyViewer, PropertyPlotter, GlobalPropertyPlotter
)
from fusion_tools.components.overlays import (
    OverlayOptions
)
from fusion_tools.components.external import (
    HRAViewer
)
from fusion_tools.components.extractors import (
    DataExtractor
)
from fusion_tools.components.custom import (
    CustomFunction, FUSIONFunction
)
from fusion_tools.components.annotation import (
    AnnotationSchema, FeatureAnnotation, BulkLabels, SlideAnnotation
)




"""

Making some utility functions for handling --omics data

"""

import os
import sys

import numpy as np
import pandas as pd

from copy import deepcopy
import requests
from typing_extensions import Union

from fusion_tools.utils.shapes import spatially_aggregate

INFO_URL = 'https://mygene.info/v3/'
HRA_URL = 'https://grlc.io/api-git/hubmapconsortium/ccf-grlc/subdir/fusion//?endpoint=https://lod.humanatlas.io/sparql'

def get_gene_info(id:Union[str,list],species: str = 'human', fields:list = ['HGNC','alias','summary'],size:int=5):
    """
    Get information about a given gene id or list of ids.
    By default returns HGNC, alias, and summary
    Can be expanded to include go, pubmed articles, etc.
    """
    acceptable_species = ['human','mouse','rat','fruitfly','nematode','zebrafish','thale-cress','frog','pig']
    assert species in acceptable_species

    # There are actually too many possible fields to mention here
    #acceptable_fields = ['HGNC','alias','summary','go','pubmed']
    #assert all([i in acceptable_fields for i in fields])
    if isinstance(id,str):
        id = id.split('.')[0]
        request_response = requests.get(f'{INFO_URL}gene/{id}?fields={",".join(fields)}&species={species}&dotfield=false&size={size}')
        if request_response.ok:
            return request_response.json()
        else:
            return None
        
    elif isinstance(id,list):
        return_list = []
        for i in id:
            request_response = requests.get(f'{INFO_URL}gene/{i.split(".")[0]}?fields={",".join(fields)}&species={species}&dotfield=false&size={size}')
            if request_response.ok:
                return_list.append(request_response.json())
        
        return return_list

def get_asct(id:Union[str,int]):
    """
    Get Anatomical Structure & Cell Type associated with a given HGNC Id.
    """
    # Have to add on the whole iri here:
    # HGNC id is a number but should be interpreted as a string
    id = f'http://identifiers.org/hgnc/{id}'
    request_response = requests.get(
        f'{HRA_URL.replace("fusion//","fusion//asct_by_biomarker")}&biomarker={id}',
        headers={'Accept':'application/json','Content-Type':'application/json'}
    )
    if request_response.ok:
        return pd.json_normalize(request_response.json()['results']['bindings'],max_level=1)
    else:
        print(f'{HRA_URL.replace("fusion//","fusion//asct_by_biomarker")}&biomarker={id}')
        print('Request not ok!')
        return None
    
def get_cell(id:str):
    """
    Get all the cell types available within a given anatomical structure
    Input has to be an UBERON id "UBERON_######..."
    """

    # Modifiying input id
    id = f'http://purl.odolibrary.org/obo/{id}'
    request_response = requests.get(
        f'{HRA_URL.replace("fusion//","fusion//cell_by_location")}&location={id}',
        headers={'Accept':'application/json','Content-type':'application/json'}
    )

    if request_response.ok:
        return pd.DataFrame(request_response.content)
    else:
        return None

def selective_aggregation(child_geo:dict, parent_geo:dict, include_keys: dict = {}, aggregate_dropped: bool = True, dropped_name: str = 'undefined', re_normalize: bool = True):
    """Selectively aggregate different fields for each intersecting structure. Useful for only aggregating cell types which should be found within a specific structure

    :param child_geo: Child structure to be receiving aggregated properties
    :type child_geo: dict
    :param parent_geo: Parent structure that contains properties that are going to be aggregated within the child geos.
    :type parent_geo: dict
    :param include_keys: Dictionary containing keys for each property that is modified and a list of values which should be included for that key., defaults to {}
    :type include_keys: dict, optional
    :param aggregate_dropped: Whether or not to include dropped values in the final aggregation, defaults to True
    :type aggregate_dropped: bool, optional
    :param dropped_name: Name to use for dropped names that are aggregated. Ignored if aggregate_dropped=False, defaults to 'undefined'
    :type dropped_name: str, optional
    :param re_normalize: Whether or not to re-normalize values after dropping keys., defaults to True
    :type re_normalize: bool, optional
    :return: Child structure with selectively aggregated properties from the parent structure.
    :rtype: dict
    """

    mod_parent = deepcopy(parent_geo)
    for f_idx,f in enumerate(parent_geo['features']):
        for k,v in include_keys.items():
            if k in f['properties']:
                sum_f_k = sum(list(f['properties'][k].values()))

                mod_parent['features'][f_idx]['properties'][k] = {
                    i: f['properties'][i]
                    for i in v
                }
                if aggregate_dropped:
                    mod_parent['features'][f_idx]['properties'][k][dropped_name] = sum_f_k - sum(list(mod_parent['features'][f_idx]['properties'][k].values()))
                if re_normalize:
                    k_sum = sum(list(mod_parent['features'][f_idx]['properties'][k].values()))
                    if k_sum>0:
                        mod_parent['features'][f_idx]['properties'][k] = {
                            i: j / k_sum
                            for i,j in mod_parent['features'][f_idx]['properties'][k].items()
                        }
    
    aggregated_child = spatially_aggregate(child_geo,[mod_parent],separate = False, summarize = False)
    
    return aggregated_child

def group_subtypes(geo_props: dict, name: str, key: dict, keep_zeros: bool = True, normalize: bool = True)->dict:
    """Grouping together properties into an lower-level descriptor

    :param geo_props: Property dict for a single Feature
    :type geo_props: dict
    :param name: Name of property containing "sub-properties" to be aggregated
    :type name: str
    :param key: Dictionary containing keys and values pertaining to sub-properties to be aggregated to each key
    :type key: dict
    :param keep_zeros: Whether or not to keep aggregated properties which sum to zero, defaults to True
    :type keep_zeros: bool, optional
    :param normalize: Whether to normalize this set of keys to sum to 1 or keep as sums, defaults to True
    :type normalize: bool, optional
    :return: Updated property dictionary containing lower-level descriptor and keys
    :rtype: dict
    """
    return_dict = {}
    if name in geo_props:
        sub_props = geo_props[name]
        for k,v in key.items():
            if type(v)==list:
                v_vals = [sub_props[v_i] if v_i in sub_props else 0 for v_i in v]
            elif type(v)==str:
                v_vals = [sub_props[v] if v in sub_props else 0]
            
            if sum(v_vals)==0:
                if keep_zeros:
                    return_dict[k] = 0.0
            else:
                return_dict[k] = sum(v_vals)
        
        if normalize:
            dict_sum = sum(list(return_dict.values()))
            if dict_sum>0:
                norm_dict = {
                    i: j/dict_sum
                    for i,j in return_dict.items()
                }
            else:
                norm_dict = return_dict.copy()

            return norm_dict
        else:
            return return_dict
    else:
        return return_dict


